# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.kernel as pmo
from pyomo.core import ConcreteModel, Var, Objective, Constraint, maximize
from pyomo.opt import TerminationCondition
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


@register_model
class LP_infeasible2(_BaseTestModel):
    """
    An infeasible LP
    """

    description = "LP_infeasible2"
    capabilities = set(['linear'])
    test_pickling = False

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.solve_should_fail = True
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.x = Var(bounds=(1, None))
        model.y = Var(bounds=(1, None))
        model.o = Objective(expr=-model.x - model.y, sense=maximize)
        model.c = Constraint(expr=model.x + model.y <= 0)

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        model.x.value = None
        model.y.value = None

    def post_solve_test_validation(self, tester, results):
        if tester is None:
            assert (
                results['Solver'][0]['termination condition']
                == TerminationCondition.infeasible
            )
        else:
            tester.assertEqual(
                results['Solver'][0]['termination condition'],
                TerminationCondition.infeasible,
            )


@register_model
class LP_infeasible2_kernel(LP_infeasible2):
    def _generate_model(self):
        self.model = pmo.block()
        model = self.model
        model._name = self.description

        model.x = pmo.variable(lb=1)
        model.y = pmo.variable(lb=1)
        model.o = pmo.objective(-model.x - model.y, sense=pmo.maximize)
        model.c = pmo.constraint(model.x + model.y <= 0)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.core.base import (
    Transformation,
    TransformationFactory,
    Constraint,
    Var,
    Block,
    ComponentUID,
    SortComponents,
    value,
)
from pyomo.mpec.complementarity import Complementarity
from pyomo.gdp import Disjunct

logger = logging.getLogger('pyomo.core')


#
# This transformation reworks each Complementarity block to
# create a mixed-complementarity problem that can be written to an NL file.
#
@TransformationFactory.register(
    'mpec.nl', doc="Transform a MPEC into a form suitable for the NL writer"
)
class MPEC4_Transformation(Transformation):
    def __init__(self):
        super(MPEC4_Transformation, self).__init__()

    def _apply_to(self, instance, **kwds):
        #
        # Find the free variables
        #
        free_vars = {}
        id_list = []
        # [ESJ 07/12/2019] Look on the whole model in case instance is a Block or a Disjunct
        for vdata in instance.model().component_data_objects(
            Var,
            active=True,
            sort=SortComponents.deterministic,
            descend_into=(Block, Disjunct),
        ):
            id_list.append(id(vdata))
            free_vars[id(vdata)] = vdata
        #
        # Iterate over the Complementarity components
        #
        cobjs = []
        for cobj in instance.component_objects(
            Complementarity,
            active=True,
            descend_into=(Block, Disjunct),
            sort=SortComponents.deterministic,
        ):
            cobjs.append(cobj)
            for index in sorted(cobj.keys()):
                _cdata = cobj[index]
                if not _cdata.active:
                    continue
                #
                # Apply a variant of the standard form logic
                #
                self.to_common_form(_cdata, free_vars)
                #
        tdata = instance._transformation_data['mpec.nl']
        tdata.compl_cuids = []
        for cobj in cobjs:
            tdata.compl_cuids.append(ComponentUID(cobj))
            cobj.parent_block().reclassify_component_type(cobj, Block)

    # instance.pprint()
    # self.print_nl_form(instance)

    def print_nl_form(self, instance):  # pragma:nocover
        """
        Summarize the complementarity relations in this problem.
        """
        vmap = {}
        for vdata in instance.component_data_objects(Var, active=True):
            vmap[id(vdata)] = vdata
        print("-------------------- Complementary Relations ----------------------")
        for bdata in instance.block_data_objects(
            active=True, sort=SortComponents.deterministic
        ):
            for cobj in bdata.component_data_objects(
                Constraint, active=True, descend_into=False
            ):
                print(
                    "%s %s\t\t\t%s"
                    % (
                        getattr(cobj, '_complementarity', None),
                        str(cobj.lower)
                        + " < "
                        + str(cobj.body)
                        + " < "
                        + str(cobj.upper),
                        vmap.get(getattr(cobj, '_vid', None), None),
                    )
                )
        print("-------------------- Complementary Relations ----------------------")

    def to_common_form(self, cdata, free_vars):
        """
        Convert a common form that can processed by AMPL
        """
        _e1 = cdata._canonical_expression(cdata._args[0])
        _e2 = cdata._canonical_expression(cdata._args[1])
        if False:  # pragma:nocover
            if _e1[0] is None:
                print(None)
            else:
                print(str(_e1[0]))
            if _e1[1] is None:
                print(None)
            else:
                print(str(_e1[1]))
            if len(_e1) > 2:
                if _e1[2] is None:
                    print(None)
                else:
                    print(str(_e1[2]))
            if _e2[0] is None:
                print(None)
            else:
                print(str(_e2[0]))
            if _e2[1] is None:
                print(None)
            else:
                print(str(_e2[1]))
            if len(_e2) > 2:
                if _e2[2] is None:
                    print(None)
                else:
                    print(str(_e2[2]))
        if len(_e1) == 2:
            cdata.c = Constraint(expr=_e1)
            return
        if len(_e2) == 2:
            cdata.c = Constraint(expr=_e2)
            return
        if (_e1[0] is None) + (_e1[2] is None) + (_e2[0] is None) + (
            _e2[2] is None
        ) != 2:
            raise RuntimeError(
                "Complementarity condition %s must have exactly two finite bounds"
                % cdata.name
            )
        #
        # Swap if the body of the second constraint is not a free variable
        #
        if not id(_e2[1]) in free_vars and id(_e1[1]) in free_vars:
            _e1, _e2 = _e2, _e1
        #
        # Rework the first constraint to have a zero bound.
        # The bound is a lower or upper bound depending on the
        # variable bound.
        #
        if not _e1[0] is None:
            cdata.bv = Var()
            cdata.c = Constraint(expr=0 <= cdata.bv)
            if not _e2[0] is None:
                cdata.bc = Constraint(expr=cdata.bv == _e1[1] - _e1[0])
            else:
                cdata.bc = Constraint(expr=cdata.bv == _e1[0] - _e1[1])
        elif not _e1[2] is None:
            cdata.bv = Var()
            cdata.c = Constraint(expr=0 <= cdata.bv)
            if not _e2[2] is None:
                cdata.bc = Constraint(expr=cdata.bv == _e1[1] - _e1[2])
            else:
                cdata.bc = Constraint(expr=cdata.bv == _e1[2] - _e1[1])
        else:
            cdata.bv = Var()
            cdata.bc = Constraint(expr=cdata.bv == _e1[1])
            cdata.c = Constraint(expr=(None, cdata.bv, None))
        #
        # If the body of the second constraint is a free variable, then keep it.
        # Otherwise, create a new variable and a new constraint.
        #
        if id(_e2[1]) in free_vars:
            var = _e2[1]
            cdata.c._vid = id(_e2[1])
            del free_vars[cdata.c._vid]
        else:
            var = cdata.v = Var()
            cdata.c._vid = id(cdata.v)
            cdata.e = Constraint(expr=cdata.v == _e2[1])
        #
        # Set the variable bound values, and corresponding _complementarity value
        #
        cdata.c._complementarity = 0
        if not _e2[0] is None:
            if var.lb is None or value(_e2[0]) > value(var.lb):
                var.setlb(_e2[0])
            cdata.c._complementarity += 1
        if not _e2[2] is None:
            if var.ub is None or value(_e2[2]) > value(var.ub):
                var.setub(_e2[2])
            cdata.c._complementarity += 2



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr import ProductExpression, PowExpression
from pyomo.core.expr.numvalue import as_numeric
from pyomo.core import Binary, value
from pyomo.core.base import (
    Transformation,
    TransformationFactory,
    Var,
    Constraint,
    ConstraintList,
    Block,
    RangeSet,
)
from pyomo.core.base.var import VarData

import logging

logger = logging.getLogger(__name__)


@TransformationFactory.register(
    "core.radix_linearization",
    doc="Linearize bilinear and quadratic terms through "
    "radix discretization (multiparametric disaggregation)",
)
class RadixLinearization(Transformation):
    """
    This plugin generates linear relaxations of bilinear problems using
    the multiparametric disaggregation technique of Kolodziej, Castro,
    and Grossmann.  See:

    Scott Kolodziej, Pedro M. Castro, and Ignacio E. Grossmann. "Global
       optimization of bilinear programs with a multiparametric
       disaggregation technique."  J.Glob.Optim 57 pp.1039-1063. 2013.
       (DOI 10.1007/s10898-012-0022-1)
    """

    def _create_using(self, model, **kwds):
        precision = kwds.pop('precision', 8)
        user_discretize = kwds.pop('discretize', set())
        verbose = kwds.pop('verbose', False)

        M = model.clone()

        # TODO: if discretize is not empty, we must translate those
        # components over to the components on the cloned instance
        _discretize = {}
        if user_discretize:
            for _var in user_discretize:
                _v = M.find_component(_var.name)
                if _v.component() is _v:
                    for _vv in _v.itervalues():
                        _discretize.setdefault(id(_vv), len(_discretize))
                else:
                    _discretize.setdefault(id(_v), len(_discretize))

        # Iterate over all Constraints and identify the bilinear and
        # quadratic terms
        bilinear_terms = []
        quadratic_terms = []
        for constraint in M.component_map(Constraint, active=True).itervalues():
            for cname, c in constraint._data.iteritems():
                if c.body.polynomial_degree() != 2:
                    continue
                self._collect_bilinear(c.body, bilinear_terms, quadratic_terms)

        # We want to find the (minimum?) number of variables to
        # discretize so that we cover all the bilinearities -- without
        # discretizing both sides of any single bilinear expression.
        # First step: figure out how many expressions each term appears
        # in
        _counts = {}
        for q in quadratic_terms:
            if not q[1].is_continuous():
                continue
            _id = id(q[1])
            if _id not in _counts:
                _counts[_id] = (q[1], set())
            _counts[_id][1].add(_id)
        for bi in bilinear_terms:
            for i in (0, 1):
                if not bi[i + 1].is_continuous():
                    continue
                _id = id(bi[i + 1])
                if _id not in _counts:
                    _counts[_id] = (bi[i + 1], set())
                _counts[_id][1].add(id(bi[2 - i]))

        _tmp_counts = dict(_counts)
        # First, remove the variables that the user wants to have discretized
        for _id in _discretize:
            for _i in _tmp_counts[_id][1]:
                if _i == _id:
                    continue
                _tmp_counts[_i][1].remove(_id)
            del _tmp_counts[_id]
        # All quadratic terms must be discretized (?)
        # for q in quadratic_terms:
        #    _id = id(q[1])
        #    if _id not in _tmp_counts:
        #        continue
        #    _discretize.setdefault(_id, len(_discretize))
        #    for _i in _tmp_counts[_id][1]:
        #        if _i == _id:
        #            continue
        #        _tmp_counts[_i][1].remove(_id)
        #    del _tmp_counts[_id]

        # Now pick a (minimal) subset of the terms in bilinear expressions
        while _tmp_counts:
            _ct, _id = max((len(_tmp_counts[i][1]), i) for i in _tmp_counts)
            if not _ct:
                break
            _discretize.setdefault(_id, len(_discretize))
            for _i in list(_tmp_counts[_id][1]):
                if _i == _id:
                    continue
                _tmp_counts[_i][1].remove(_id)
            del _tmp_counts[_id]

        #
        # Discretize things
        #

        # Define a block (namespace) for holding the disaggregated
        # variables and new constraints
        if False:  # Set to true when the LP writer is fixed
            M._radix_linearization = Block()
            _block = M._radix_linearization
        else:
            _block = M
        _block.DISCRETIZATION = RangeSet(precision)
        _block.DISCRETIZED_VARIABLES = RangeSet(0, len(_discretize) - 1)
        _block.z = Var(
            _block.DISCRETIZED_VARIABLES, _block.DISCRETIZATION, within=Binary
        )
        _block.dv = Var(_block.DISCRETIZED_VARIABLES, bounds=(0, 2**-precision))

        # Actually discretize the terms we have marked for discretization
        for _id, _idx in _discretize.items():
            if verbose:
                logger.info(
                    "Discretizing variable %s as %s" % (_counts[_id][0].name, _idx)
                )
            self._discretize_variable(_block, _counts[_id][0], _idx)

        _known_bilinear = {}
        # For each quadratic term, if it hasn't been discretized /
        # generated, do so, and remember the resulting W term for later
        # use...
        # for _expr, _x1 in quadratic_terms:
        #    self._discretize_term( _expr, _x1, _x1,
        #                           _block, _discretize, _known_bilinear )
        # For each bilinear term, if it hasn't been discretized /
        # generated, do so, and remember the resulting W term for later
        # use...
        for _expr, _x1, _x2 in bilinear_terms:
            self._discretize_term(_expr, _x1, _x2, _block, _discretize, _known_bilinear)

        # Return the discretized instance!
        return M

    def _discretize_variable(self, b, v, idx):
        _lb, _ub = v.bounds
        if _lb is None or _ub is None:
            raise RuntimeError(
                "Couldn't discretize variable %s: missing "
                "finite lower/upper bounds." % (v.name)
            )
        _c = Constraint(
            expr=v
            == _lb
            + (_ub - _lb)
            * (b.dv[idx] + sum(b.z[idx, k] * 2**-k for k in b.DISCRETIZATION))
        )
        b.add_component("c_discr_v%s" % idx, _c)

    def _discretize_term(self, _expr, _x1, _x2, _block, _discretize, _known_bilinear):
        if id(_x1) in _discretize:
            _v = _x1
            _u = _x2
        elif id(_x2) in _discretize:
            _u = _x1
            _v = _x2
        else:
            raise RuntimeError(
                "Couldn't identify discretized variable for expression '%s'!" % _expr
            )
        _id = (id(_v), id(_u))
        if _id not in _known_bilinear:
            _known_bilinear[_id] = self._discretize_bilinear(
                _block, _v, _discretize[id(_v)], _u, len(_known_bilinear)
            )
        # _expr should be a "simple" product expression; substitute
        # in the bilinear "W" term for the raw bilinear terms
        _expr._numerator = [_known_bilinear[_id]]

    def _discretize_bilinear(self, b, v, v_idx, u, u_idx):
        _z = b.z
        _dv = b.dv[v_idx]
        _u = Var(b.DISCRETIZATION, within=u.domain, bounds=u.bounds)
        logger.info(
            "Discretizing (v=%s)*(u=%s) as u%s_v%s" % (v.name, u.name, u_idx, v_idx)
        )
        b.add_component("u%s_v%s" % (u_idx, v_idx), _u)
        _lb, _ub = u.bounds
        if _lb is None or _ub is None:
            raise RuntimeError(
                "Couldn't relax variable %s: missing "
                "finite lower/upper bounds." % (u.name)
            )
        _c = ConstraintList()
        b.add_component("c_disaggregate_u%s_v%s" % (u_idx, v_idx), _c)
        for k in b.DISCRETIZATION:
            # _lb * z[v_idx,k] <= _u[k] <= _ub * z[v_idx,k]
            _c.add(expr=_lb * _z[v_idx, k] <= _u[k])
            _c.add(expr=_u[k] <= _ub * _z[v_idx, k])
            # _lb * (1-z[v_idx,k]) <= u - _u[k] <= _ub * (1-z[v_idx,k])
            _c.add(expr=_lb * (1 - _z[v_idx, k]) <= u - _u[k])
            _c.add(expr=u - _u[k] <= _ub * (1 - _z[v_idx, k]))

        _v_lb, _v_ub = v.bounds
        _bnd_rng = (_v_lb * _lb, _v_lb * _ub, _v_ub * _lb, _v_ub * _ub)
        _w = Var(bounds=(min(_bnd_rng), max(_bnd_rng)))
        b.add_component("w%s_v%s" % (u_idx, v_idx), _w)

        K = max(b.DISCRETIZATION)

        _dw = Var(
            bounds=(min(0, _lb * 2**-K, _ub * 2**-K), max(0, _lb * 2**-K, _ub * 2**-K))
        )
        b.add_component("dw%s_v%s" % (u_idx, v_idx), _dw)

        _c = Constraint(
            expr=_w
            == _v_lb * u
            + (_v_ub - _v_lb) * (sum(2**-k * _u[k] for k in b.DISCRETIZATION) + _dw)
        )
        b.add_component("c_bilinear_u%s_v%s" % (u_idx, v_idx), _c)

        _c = ConstraintList()
        b.add_component("c_mccormick_u%s_v%s" % (u_idx, v_idx), _c)
        # u_lb * dv <= dw <= u_ub * dv
        _c.add(expr=_lb * _dv <= _dw)
        _c.add(expr=_dw <= _ub * _dv)
        # (u-u_ub)*2^-K + u_ub*dv <= dw <= (u-u_lb)*2^-K + u_lb*dv
        _c.add(expr=(u - _ub) * 2**-K + _ub * _dv <= _dw)
        _c.add(expr=_dw <= (u - _lb) * 2**-K + _lb * _dv)

        return _w

    def _collect_bilinear(self, expr, bilin, quad):
        if not expr.is_expression_type():
            return
        if type(expr) is ProductExpression:
            if len(expr._numerator) != 2:
                for e in expr._numerator:
                    self._collect_bilinear(e, bilin, quad)
                # No need to check denominator, as this is poly_degree==2
                return
            if not isinstance(expr._numerator[0], VarData) or not isinstance(
                expr._numerator[1], VarData
            ):
                raise RuntimeError("Cannot yet handle complex subexpressions")
            if expr._numerator[0] is expr._numerator[1]:
                quad.append((expr, expr._numerator[0]))
            else:
                bilin.append((expr, expr._numerator[0], expr._numerator[1]))
            return
        if type(expr) is PowExpression and value(expr._args[1]) == 2:
            # Note: directly testing the value of the exponent above is
            # safe: we have already verified that this expression is
            # polynomial, so the exponent must be constant.
            tmp = ProductExpression()
            tmp._numerator = [expr._args[0], expr._args[0]]
            tmp._denominator = []
            expr._args = (tmp, as_numeric(1))  # THIS CODE DOES NOT WORK
            # quad.append( (tmp, tmp._args[0]) )
            self._collect_bilinear(tmp, bilin, quad)
            return
        # All other expression types
        for e in expr._args:
            self._collect_bilinear(e, bilin, quad)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests explicit bound to variable bound transformation module."""

import pyomo.common.unittest as unittest
from pyomo.contrib.gdp_bounds.info import disjunctive_lb, disjunctive_ub
from pyomo.environ import (
    ConcreteModel,
    Constraint,
    Objective,
    TransformationFactory,
    Var,
)
from pyomo.gdp import Disjunct, Disjunction
from pyomo.opt import check_available_solvers

solvers = check_available_solvers('cbc')


class TestGDPBounds(unittest.TestCase):
    """Tests disjunctive variable bounds implementation."""

    @unittest.skipIf('cbc' not in solvers, "CBC solver not available")
    def test_compute_bounds_obbt(self):
        """Test computation of disjunctive bounds."""
        m = ConcreteModel()
        m.x = Var(bounds=(0, 8))
        m.d1 = Disjunct()
        m.d1.c = Constraint(expr=m.x >= 2)
        m.d2 = Disjunct()
        m.d2.c = Constraint(expr=m.x <= 4)
        m.disj = Disjunction(expr=[m.d1, m.d2])
        m.obj = Objective(expr=m.x)
        TransformationFactory('contrib.compute_disj_var_bounds').apply_to(
            m, solver='cbc'
        )
        self.assertEqual(m.d1._disj_var_bounds[m.x], (2, 8))
        self.assertEqual(m.d2._disj_var_bounds[m.x], (0, 4))
        self.assertEqual(disjunctive_lb(m.x, m.d1), 2)
        self.assertEqual(disjunctive_ub(m.x, m.d1), 8)
        self.assertEqual(disjunctive_lb(m.x, m.d2), 0)
        self.assertEqual(disjunctive_ub(m.x, m.d2), 4)

    @unittest.skipIf('cbc' not in solvers, "CBC solver not available")
    def test_compute_bounds_obbt_prune_disjunct(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 1))
        m.d1 = Disjunct()
        m.d1.c = Constraint(expr=m.x >= 2)
        m.d1.c2 = Constraint(expr=m.x <= 1)
        m.d2 = Disjunct()
        m.d2.c = Constraint(expr=m.x + 3 == 0)
        m.disj = Disjunction(expr=[m.d1, m.d2])
        m.obj = Objective(expr=m.x)
        TransformationFactory('contrib.compute_disj_var_bounds').apply_to(
            m, solver='cbc'
        )
        self.assertFalse(m.d1.active)
        self.assertEqual(m.d1.binary_indicator_var.value, 0)
        self.assertTrue(m.d1.indicator_var.fixed)

    def test_compute_bounds_fbbt(self):
        """Test computation of disjunctive bounds."""
        m = ConcreteModel()
        m.x = Var(bounds=(0, 8))
        m.d1 = Disjunct()
        m.d1.c = Constraint(expr=m.x >= 2)
        m.d2 = Disjunct()
        m.d2.c = Constraint(expr=m.x <= 4)
        m.disj = Disjunction(expr=[m.d1, m.d2])
        m.obj = Objective(expr=m.x)
        TransformationFactory('contrib.compute_disj_var_bounds').apply_to(m)
        self.assertEqual(m.d1._disj_var_bounds[m.x], (2, 8))
        self.assertEqual(m.d2._disj_var_bounds[m.x], (0, 4))
        self.assertEqual(disjunctive_lb(m.x, m.d1), 2)
        self.assertEqual(disjunctive_ub(m.x, m.d1), 8)
        self.assertEqual(disjunctive_lb(m.x, m.d2), 0)
        self.assertEqual(disjunctive_ub(m.x, m.d2), 4)

    def test_nested_fbbt(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 8))
        m.y = Var(bounds=(0, 8))
        m.d1 = Disjunct()
        m.d1.c = Constraint(expr=m.x >= 2)
        m.d1.innerD1 = Disjunct()
        m.d1.innerD1.c = Constraint(expr=m.y >= m.x + 3)
        m.d1.innerD2 = Disjunct()
        m.d1.innerD2.c = Constraint(expr=m.y <= m.x - 4)
        m.d1.innerDisj = Disjunction(expr=[m.d1.innerD1, m.d1.innerD2])
        m.d2 = Disjunct()
        m.d2.c = Constraint(expr=m.x == 3)
        m.disj = Disjunction(expr=[m.d1, m.d2])
        TransformationFactory('contrib.compute_disj_var_bounds').apply_to(m)
        self.assertEqual(disjunctive_lb(m.y, m.d1), 0)
        self.assertEqual(disjunctive_lb(m.y, m.d1.innerD1), 5)
        self.assertEqual(disjunctive_ub(m.y, m.d1.innerD2), 4)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.contrib.gdpopt.enumerate import GDP_Enumeration_Solver

from pyomo.environ import (
    SolverFactory,
    Objective,
    maximize,
    TerminationCondition,
    value,
    Var,
    Integers,
    Constraint,
    ConcreteModel,
)
from pyomo.gdp import Disjunction
import pyomo.gdp.tests.models as models


@unittest.skipUnless(SolverFactory('gurobi').available(), 'Gurobi not available')
@unittest.skipUnless(SolverFactory('gurobi').license_is_valid(), 'Gurobi not licensed')
class TestGDPoptEnumerate(unittest.TestCase):
    def test_solve_two_term_disjunction(self):
        m = models.makeTwoTermDisj()
        m.obj = Objective(expr=m.x, sense=maximize)

        results = SolverFactory('gdpopt.enumerate').solve(m)

        self.assertEqual(results.solver.iterations, 2)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertEqual(results.problem.lower_bound, 9)
        self.assertEqual(results.problem.upper_bound, 9)

        self.assertEqual(value(m.x), 9)
        self.assertTrue(value(m.d[0].indicator_var))
        self.assertFalse(value(m.d[1].indicator_var))

    def modify_two_term_disjunction(self, m):
        # Make first disjunct feasible
        m.a.setlb(0)
        # Discrete variable
        m.y = Var(domain=Integers, bounds=(2, 4))
        m.d[1].c3 = Constraint(expr=m.x <= 6)
        m.d[0].c2 = Constraint(expr=m.y + m.a - 5 <= 2)

        m.obj = Objective(expr=-m.x - m.y)

    def test_solve_GDP_iterate_over_discrete_variables(self):
        m = models.makeTwoTermDisj()
        self.modify_two_term_disjunction(m)

        results = SolverFactory('gdpopt.enumerate').solve(m, force_subproblem_nlp=True)

        self.assertEqual(results.solver.iterations, 6)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertEqual(results.problem.lower_bound, -11)
        self.assertEqual(results.problem.upper_bound, -11)

        self.assertEqual(value(m.x), 9)
        self.assertEqual(value(m.y), 2)
        self.assertTrue(value(m.d[0].indicator_var))
        self.assertFalse(value(m.d[1].indicator_var))

    def test_solve_GDP_do_not_iterate_over_discrete_variables(self):
        m = models.makeTwoTermDisj()
        self.modify_two_term_disjunction(m)

        results = SolverFactory('gdpopt.enumerate').solve(m)

        self.assertEqual(results.solver.iterations, 2)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertEqual(results.problem.lower_bound, -11)
        self.assertEqual(results.problem.upper_bound, -11)

        self.assertEqual(value(m.x), 9)
        self.assertEqual(value(m.y), 2)
        self.assertTrue(value(m.d[0].indicator_var))
        self.assertFalse(value(m.d[1].indicator_var))

    def test_solve_GDP_iterate_over_Boolean_variables(self):
        m = models.makeLogicalConstraintsOnDisjuncts()

        results = SolverFactory('gdpopt.enumerate').solve(m, force_subproblem_nlp=True)

        self.assertEqual(results.solver.iterations, 16)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertEqual(results.problem.lower_bound, 8)
        self.assertEqual(results.problem.upper_bound, 8)

        self.assertTrue(value(m.d[2].indicator_var))
        self.assertTrue(value(m.d[3].indicator_var))
        self.assertFalse(value(m.d[1].indicator_var))
        self.assertFalse(value(m.d[4].indicator_var))
        self.assertEqual(value(m.x), 8)
        # We don't know what values they take, but they have to be different
        self.assertNotEqual(value(m.Y[1]), value(m.Y[2]))

    def test_solve_GDP_do_not_iterate_over_Boolean_variables(self):
        m = models.makeLogicalConstraintsOnDisjuncts()

        results = SolverFactory('gdpopt.enumerate').solve(m)

        self.assertEqual(results.solver.iterations, 4)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertEqual(results.problem.lower_bound, 8)
        self.assertEqual(results.problem.upper_bound, 8)

        self.assertTrue(value(m.d[2].indicator_var))
        self.assertTrue(value(m.d[3].indicator_var))
        self.assertFalse(value(m.d[1].indicator_var))
        self.assertFalse(value(m.d[4].indicator_var))
        self.assertEqual(value(m.x), 8)
        # We don't know what values they take, but they have to be different
        self.assertNotEqual(value(m.Y[1]), value(m.Y[2]))

    def test_stop_at_iteration_limit(self):
        m = models.makeLogicalConstraintsOnDisjuncts()

        results = SolverFactory('gdpopt.enumerate').solve(
            m, iterlim=4, force_subproblem_nlp=True
        )

        self.assertEqual(results.solver.iterations, 4)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.maxIterations
        )

    def test_unbounded_GDP(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-1, 10))
        m.y = Var(bounds=(2, 3))
        m.z = Var()
        m.d = Disjunction(expr=[[m.x + m.y >= 5], [m.x - m.y <= 3]])
        m.o = Objective(expr=m.z)

        results = SolverFactory('gdpopt.enumerate').solve(m)

        self.assertEqual(results.solver.iterations, 1)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.unbounded
        )
        self.assertEqual(results.problem.lower_bound, -float('inf'))
        self.assertEqual(results.problem.upper_bound, -float('inf'))


@unittest.skipUnless(SolverFactory('ipopt').available(), 'Ipopt not available')
class TestGDPoptEnumerate_ipopt_tests(unittest.TestCase):
    def test_infeasible_GDP(self):
        m = models.make_infeasible_gdp_model()

        results = SolverFactory('gdpopt.enumerate').solve(m)

        self.assertEqual(results.solver.iterations, 2)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.infeasible
        )
        self.assertEqual(results.problem.lower_bound, float('inf'))

    def test_algorithm_specified_to_solve(self):
        m = models.twoDisj_twoCircles_easy()

        results = SolverFactory('gdpopt').solve(m, algorithm='enumerate', tee=True)

        self.assertEqual(results.solver.iterations, 2)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertAlmostEqual(results.problem.lower_bound, 9)
        self.assertAlmostEqual(results.problem.upper_bound, 9)

        self.assertAlmostEqual(value(m.x), 2)
        self.assertAlmostEqual(value(m.y), 7)
        self.assertTrue(value(m.upper_circle.indicator_var))
        self.assertFalse(value(m.lower_circle.indicator_var))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from io import StringIO
import logging
import pickle

from pyomo.common.dependencies import attempt_import
from pyomo.common.log import LoggingIntercept
import pyomo.common.unittest as unittest
from pyomo.contrib.piecewise import PiecewiseLinearFunction, Triangulation
from pyomo.core.expr.compare import (
    assertExpressionsEqual,
    assertExpressionsStructurallyEqual,
)
from pyomo.environ import ConcreteModel, Constraint, log, Var

np, numpy_available = attempt_import('numpy')
scipy, scipy_available = attempt_import('scipy')


def f(x):
    return log(x)


def f1(x):
    return (log(3) / 2) * x - log(3) / 2


def f2(x):
    return (log(2) / 3) * x + log(3 / 2)


def f3(x):
    return (log(5 / 3) / 4) * x + log(6 / ((5 / 3) ** (3 / 2)))


class TestPiecewiseLinearFunction2D(unittest.TestCase):
    def make_ln_x_model(self):
        m = ConcreteModel()
        m.x = Var(bounds=(1, 10))
        m.f = f
        m.f1 = f1
        m.f2 = f2
        m.f3 = f3

        return m

    def check_ln_x_approx(self, pw, x):
        self.assertEqual(len(pw._simplices), 3)
        self.assertEqual(len(pw._linear_functions), 3)
        # indices of extreme points.
        simplices = [(0, 1), (1, 2), (2, 3)]
        for idx, simplex in enumerate(simplices):
            self.assertEqual(pw._simplices[idx], simplices[idx])

        assertExpressionsEqual(
            self, pw._linear_functions[0](x), (log(3) / 2) * x - log(3) / 2, places=7
        )
        assertExpressionsEqual(
            self, pw._linear_functions[1](x), (log(2) / 3) * x + log(3 / 2), places=7
        )
        assertExpressionsEqual(
            self,
            pw._linear_functions[2](x),
            (log(5 / 3) / 4) * x + log(6 / ((5 / 3) ** (3 / 2))),
            places=7,
        )

    def check_x_squared_approx(self, pw, x):
        self.assertEqual(len(pw._simplices), 3)
        self.assertEqual(len(pw._linear_functions), 3)
        # indices of extreme points.
        simplices = [(0, 1), (1, 2), (2, 3)]
        for idx, simplex in enumerate(simplices):
            self.assertEqual(pw._simplices[idx], simplices[idx])

        assertExpressionsStructurallyEqual(
            self, pw._linear_functions[0](x), 4 * x - 3, places=7
        )
        assertExpressionsStructurallyEqual(
            self, pw._linear_functions[1](x), 9 * x - 18, places=7
        )
        assertExpressionsStructurallyEqual(
            self, pw._linear_functions[2](x), 16 * x - 60, places=7
        )

    def test_pw_linear_approx_of_ln_x_simplices(self):
        m = self.make_ln_x_model()
        simplices = [(1, 3), (3, 6), (6, 10)]
        m.pw = PiecewiseLinearFunction(simplices=simplices, function=m.f)
        self.check_ln_x_approx(m.pw, m.x)

    def test_pw_linear_approx_of_ln_x_points(self):
        m = self.make_ln_x_model()
        m.pw = PiecewiseLinearFunction(points=[1, 3, 6, 10], function=m.f)
        self.check_ln_x_approx(m.pw, m.x)

    def test_pw_linear_approx_of_ln_x_linear_funcs(self):
        m = self.make_ln_x_model()
        m.pw = PiecewiseLinearFunction(
            simplices=[(1, 3), (3, 6), (6, 10)], linear_functions=[m.f1, m.f2, m.f3]
        )
        self.check_ln_x_approx(m.pw, m.x)

    def test_pw_linear_approx_of_ln_x_tabular_data(self):
        m = self.make_ln_x_model()
        m.pw = PiecewiseLinearFunction(
            tabular_data={1: 0, 3: log(3), 6: log(6), 10: log(10)}
        )
        self.check_ln_x_approx(m.pw, m.x)

    def test_pw_linear_approx_of_ln_x_j1(self):
        m = self.make_ln_x_model()
        m.pw = PiecewiseLinearFunction(
            points=[1, 3, 6, 10], triangulation=Triangulation.J1, function=m.f
        )
        self.check_ln_x_approx(m.pw, m.x)
        # we disregard their request because it's 1D
        self.assertEqual(m.pw.triangulation, Triangulation.AssumeValid)

    def test_pw_linear_approx_of_ln_x_user_defined_segments(self):
        m = self.make_ln_x_model()
        m.pw = PiecewiseLinearFunction(
            simplices=[[1, 3], [3, 6], [6, 10]], function=m.f
        )
        self.check_ln_x_approx(m.pw, m.x)
        self.assertEqual(m.pw.triangulation, Triangulation.Unknown)

    def test_use_pw_function_in_constraint(self):
        m = self.make_ln_x_model()
        m.pw = PiecewiseLinearFunction(
            simplices=[(1, 3), (3, 6), (6, 10)], linear_functions=[m.f1, m.f2, m.f3]
        )
        m.c = Constraint(expr=m.pw(m.x) <= 1)
        self.assertEqual(str(m.c.body.expr), "pw(x)")

    def test_evaluate_pw_function(self):
        m = self.make_ln_x_model()
        m.pw = PiecewiseLinearFunction(
            simplices=[(1, 3), (3, 6), (6, 10)], linear_functions=[m.f1, m.f2, m.f3]
        )
        self.assertAlmostEqual(m.pw(1), 0)
        self.assertAlmostEqual(m.pw(2), m.f1(2))
        self.assertAlmostEqual(m.pw(3), log(3))
        self.assertAlmostEqual(m.pw(4.5), m.f2(4.5))
        self.assertAlmostEqual(m.pw(9.2), m.f3(9.2))
        self.assertAlmostEqual(m.pw(10), log(10))

    def test_indexed_pw_linear_function_approximate_over_simplices(self):
        m = self.make_ln_x_model()
        m.z = Var([1, 2], bounds=(-10, 10))

        def g1(x):
            return x**2

        def g2(x):
            return log(x)

        m.funcs = {1: g1, 2: g2}
        simplices = [(1, 3), (3, 6), (6, 10)]
        m.pw = PiecewiseLinearFunction(
            [1, 2], simplices=simplices, function_rule=lambda m, i: m.funcs[i]
        )
        self.check_ln_x_approx(m.pw[2], m.z[2])
        self.check_x_squared_approx(m.pw[1], m.z[1])

    def test_indexed_pw_linear_function_approximate_over_points(self):
        m = self.make_ln_x_model()
        m.z = Var([1, 2], bounds=(-10, 10))

        def g1(x):
            return x**2

        def g2(x):
            return log(x)

        m.funcs = {1: g1, 2: g2}

        def silly_pts_rule(m, i):
            return [1, 3, 6, 10]

        m.pw = PiecewiseLinearFunction(
            [1, 2], points=silly_pts_rule, function_rule=lambda m, i: m.funcs[i]
        )
        self.check_ln_x_approx(m.pw[2], m.z[2])
        self.check_x_squared_approx(m.pw[1], m.z[1])

    def test_indexed_pw_linear_function_tabular_data(self):
        m = self.make_ln_x_model()
        m.z = Var([1, 2], bounds=(-10, 10))

        def silly_tabular_data_rule(m, i):
            if i == 1:
                return {1: 1, 3: 9, 6: 36, 10: 100}
            if i == 2:
                return {1: 0, 3: log(3), 6: log(6), 10: log(10)}

        m.pw = PiecewiseLinearFunction(
            [1, 2], tabular_data_rule=silly_tabular_data_rule
        )
        self.check_ln_x_approx(m.pw[2], m.z[2])
        self.check_x_squared_approx(m.pw[1], m.z[1])

    def test_indexed_pw_linear_function_linear_funcs_and_simplices(self):
        m = self.make_ln_x_model()
        m.z = Var([1, 2], bounds=(-10, 10))

        def silly_simplex_rule(m, i):
            return [(1, 3), (3, 6), (6, 10)]

        def h1(x):
            return 4 * x - 3

        def h2(x):
            return 9 * x - 18

        def h3(x):
            return 16 * x - 60

        def silly_linear_func_rule(m, i):
            return [h1, h2, h3]

        m.pw = PiecewiseLinearFunction(
            [1, 2],
            simplices=silly_simplex_rule,
            linear_functions=silly_linear_func_rule,
        )
        self.check_x_squared_approx(m.pw[1], m.z[1])
        self.check_x_squared_approx(m.pw[2], m.z[2])

    def test_pickle(self):
        m = self.make_ln_x_model()
        m.pw = PiecewiseLinearFunction(points=[1, 3, 6, 10], function=m.f)
        m.c = Constraint(expr=m.pw(m.x) >= 0.35)

        # pickle and unpickle
        unpickle = pickle.loads(pickle.dumps(m))

        # Check that the pprint is equal
        m_buf = StringIO()
        m.pprint(ostream=m_buf)
        m_output = m_buf.getvalue()

        unpickle_buf = StringIO()
        unpickle.pprint(ostream=unpickle_buf)
        unpickle_output = unpickle_buf.getvalue()
        self.assertMultiLineEqual(m_output, unpickle_output)


# Here's a cute paraboloid:
def g(x, y):
    return x**2 + y**2


class TestPiecewiseLinearFunction3D(unittest.TestCase):
    simplices = [
        [(0, 1), (0, 4), (3, 4)],
        [(0, 1), (3, 4), (3, 1)],
        [(3, 4), (3, 7), (0, 7)],
        [(0, 7), (0, 4), (3, 4)],
    ]

    def make_model(self):
        m = ConcreteModel()
        m.x1 = Var(bounds=(0, 3))
        m.x2 = Var(bounds=(1, 7))
        m.g = g
        return m

    def check_pw_linear_approximation(self, m):
        self.assertEqual(len(m.pw._simplices), 4)
        for i, simplex in enumerate(m.pw._simplices):
            for idx in simplex:
                self.assertIn(m.pw._points[idx], self.simplices[i])

        self.assertEqual(len(m.pw._linear_functions), 4)

        assertExpressionsStructurallyEqual(
            self,
            m.pw._linear_functions[0](m.x1, m.x2),
            3 * m.x1 + 5 * m.x2 - 4,
            places=7,
        )
        assertExpressionsStructurallyEqual(
            self,
            m.pw._linear_functions[1](m.x1, m.x2),
            3 * m.x1 + 5 * m.x2 - 4,
            places=7,
        )
        assertExpressionsStructurallyEqual(
            self,
            m.pw._linear_functions[2](m.x1, m.x2),
            3 * m.x1 + 11 * m.x2 - 28,
            places=7,
        )
        assertExpressionsStructurallyEqual(
            self,
            m.pw._linear_functions[3](m.x1, m.x2),
            3 * m.x1 + 11 * m.x2 - 28,
            places=7,
        )

    @unittest.skipUnless(
        scipy_available and numpy_available, "scipy and/or numpy are not available"
    )
    def test_pw_linear_approx_of_paraboloid_points(self):
        m = self.make_model()
        m.pw = PiecewiseLinearFunction(
            points=[(0, 1), (0, 4), (0, 7), (3, 1), (3, 4), (3, 7)], function=m.g
        )
        self.check_pw_linear_approximation(m)

    @unittest.skipUnless(numpy_available, "numpy is not available")
    def test_pw_linear_approx_of_paraboloid_j1(self):
        m = self.make_model()
        m.pw = PiecewiseLinearFunction(
            points=[
                (0, 1),
                (0, 4),
                (0, 7),
                (3, 1),
                (3, 4),
                (3, 7),
                (4, 1),
                (4, 4),
                (4, 7),
            ],
            function=m.g,
            triangulation=Triangulation.OrderedJ1,
        )
        self.assertEqual(len(m.pw._simplices), 8)
        self.assertEqual(m.pw.triangulation, Triangulation.OrderedJ1)

    @unittest.skipUnless(scipy_available, "scipy is not available")
    def test_pw_linear_approx_tabular_data(self):
        m = self.make_model()

        m.pw = PiecewiseLinearFunction(
            tabular_data={
                (0, 1): g(0, 1),
                (0, 4): g(0, 4),
                (0, 7): g(0, 7),
                (3, 1): g(3, 1),
                (3, 4): g(3, 4),
                (3, 7): g(3, 7),
            }
        )
        self.check_pw_linear_approximation(m)

    @unittest.skipUnless(numpy_available, "numpy are not available")
    def test_pw_linear_approx_of_paraboloid_simplices(self):
        m = self.make_model()
        m.pw = PiecewiseLinearFunction(function=m.g, simplices=self.simplices)
        self.check_pw_linear_approximation(m)

    def test_pw_linear_approx_of_paraboloid_linear_funcs(self):
        m = self.make_model()

        def g1(x1, x2):
            return 3 * x1 + 5 * x2 - 4

        def g2(x1, x2):
            return 3 * x1 + 11 * x2 - 28

        m.pw = PiecewiseLinearFunction(
            simplices=self.simplices, linear_functions=[g1, g1, g2, g2]
        )
        self.check_pw_linear_approximation(m)

    def test_use_pw_linear_approx_in_constraint(self):
        m = self.make_model()

        def g1(x1, x2):
            return 3 * x1 + 5 * x2 - 4

        def g2(x1, x2):
            return 3 * x1 + 11 * x2 - 28

        m.pw = PiecewiseLinearFunction(
            simplices=self.simplices, linear_functions=[g1, g1, g2, g2]
        )

        m.c = Constraint(expr=m.pw(m.x1, m.x2) <= 5)
        self.assertEqual(str(m.c.body.expr), "pw(x1, x2)")
        self.assertIs(m.c.body.expr.pw_linear_function, m.pw)

    @unittest.skipUnless(numpy_available, "numpy is not available")
    def test_evaluate_pw_linear_function(self):
        # NOTE: This test requires numpy because it is used to check which
        # simplex a point is in
        m = self.make_model()

        def g1(x1, x2):
            return 3 * x1 + 5 * x2 - 4

        def g2(x1, x2):
            return 3 * x1 + 11 * x2 - 28

        m.pw = PiecewiseLinearFunction(
            simplices=self.simplices, linear_functions=[g1, g1, g2, g2]
        )
        # check it's equal to the original function at all the extreme points of
        # the simplices
        for x1, x2 in m.pw._points:
            self.assertAlmostEqual(m.pw(x1, x2), m.g(x1, x2))
        # check some points in the approximation
        self.assertAlmostEqual(m.pw(1, 3), g1(1, 3))
        self.assertAlmostEqual(m.pw(2.5, 6), g2(2.5, 6))
        self.assertAlmostEqual(m.pw(0.2, 4.3), g2(0.2, 4.3))


class TestTriangulationProducesDegenerateSimplices(unittest.TestCase):
    cube_extreme_pt_indices = [
        {10, 11, 13, 14, 19, 20, 22, 23},  # right bottom back
        {9, 10, 12, 13, 18, 19, 21, 22},  # right bottom front
        {0, 1, 3, 4, 9, 10, 12, 13},  # left bottom front
        {1, 2, 4, 5, 10, 11, 13, 14},  # left bottom back
        {3, 4, 6, 7, 12, 13, 15, 16},  # left top front
        {4, 5, 7, 8, 13, 14, 16, 17},  # left top back
        {12, 13, 15, 16, 21, 22, 24, 25},  # right top front
        {13, 14, 16, 17, 22, 23, 25, 26},  # right top back
    ]

    def make_model(self):
        m = ConcreteModel()

        m.f = lambda x1, x2, y: x1 * x2 + y
        # This is a 2x2 stack of cubes, so there are 8 total cubes, each of which
        # will get divided into 6 simplices.
        m.points = [
            (-2.0, 0.0, 1.0),
            (-2.0, 0.0, 4.0),
            (-2.0, 0.0, 7.0),
            (-2.0, 1.5, 1.0),
            (-2.0, 1.5, 4.0),
            (-2.0, 1.5, 7.0),
            (-2.0, 3.0, 1.0),
            (-2.0, 3.0, 4.0),
            (-2.0, 3.0, 7.0),
            (-1.5, 0.0, 1.0),
            (-1.5, 0.0, 4.0),
            (-1.5, 0.0, 7.0),
            (-1.5, 1.5, 1.0),
            (-1.5, 1.5, 4.0),
            (-1.5, 1.5, 7.0),
            (-1.5, 3.0, 1.0),
            (-1.5, 3.0, 4.0),
            (-1.5, 3.0, 7.0),
            (-1.0, 0.0, 1.0),
            (-1.0, 0.0, 4.0),
            (-1.0, 0.0, 7.0),
            (-1.0, 1.5, 1.0),
            (-1.0, 1.5, 4.0),
            (-1.0, 1.5, 7.0),
            (-1.0, 3.0, 1.0),
            (-1.0, 3.0, 4.0),
            (-1.0, 3.0, 7.0),
        ]
        return m

    @unittest.skipUnless(
        scipy_available and numpy_available, "scipy and/or numpy are not available"
    )
    def test_degenerate_simplices_filtered(self):
        m = self.make_model()
        pw = m.approx = PiecewiseLinearFunction(points=m.points, function=m.f)

        # check that all the points got used
        self.assertEqual(len(pw._points), 27)
        for p_model, p_pw in zip(m.points, pw._points):
            self.assertEqual(p_model, p_pw)

        # Started with a 2x2 grid of cubes, and each is divided into 6
        # simplices. It's crazy degenerate in terms of *how* this is done, but
        # that's the point of this test.
        self.assertEqual(len(pw._simplices), 48)
        simplex_in_cube = {idx: 0 for idx in range(8)}
        for simplex in pw._simplices:
            for i, vertex_set in enumerate(self.cube_extreme_pt_indices):
                if set(simplex).issubset(vertex_set):
                    simplex_in_cube[i] += 1
            # verify the simplex is full-dimensional
            pts = np.array([pw._points[j] for j in simplex]).transpose()
            A = pts[:, 1:] - np.append(pts[:, :2], pts[:, [0]], axis=1)
            self.assertNotEqual(np.linalg.det(A), 0)

        # Check that they are 6 to a cube, as expected
        for num in simplex_in_cube.values():
            self.assertEqual(num, 6)

    @unittest.skipUnless(
        scipy_available and numpy_available, "scipy and/or numpy are not available"
    )
    def test_redundant_points_logged(self):
        m = self.make_model()
        # add a redundant point
        m.points.append((-2, 0, 1))

        out = StringIO()
        with LoggingIntercept(
            out, 'pyomo.contrib.piecewise.piecewise_linear_function', level=logging.INFO
        ):
            m.approx = PiecewiseLinearFunction(points=m.points, function=m.f)

        self.assertIn(
            "The Delaunay triangulation dropped the point with index 27 "
            "from the triangulation",
            out.getvalue(),
        )

    @unittest.skipUnless(numpy_available, "numpy is not available")
    def test_user_given_degenerate_simplex_error(self):
        m = self.make_model()
        with self.assertRaisesRegex(
            ValueError,
            "When calculating the hyperplane approximation over the simplex "
            "with index 0, the matrix was unexpectedly singular. This "
            "likely means that this simplex is degenerate",
        ):
            m.pw = PiecewiseLinearFunction(
                simplices=[
                    (
                        (-2.0, 0.0, 1.0),
                        (-2.0, 0.0, 4.0),
                        (-2.0, 1.5, 1.0),
                        (-2.0, 1.5, 4.0),
                    )
                ],
                function=m.f,
            )

    @unittest.skipUnless(
        scipy_available and numpy_available, "scipy and/or numpy are not available"
    )
    def test_simplex_not_numerically_full_rank_but_determinant_nonzero(self):
        m = ConcreteModel()

        def f(x3, x6, x9, x4):
            return -x6 * (0.01 * x4 * x9 + x3) + 0.98 * x3

        points = [
            (0, 0.85, 1.2, 0),
            (0.07478, 0.86396, 1.8668, 5),
            (0, 0.85, 1.8668, 0),
            (0.07478, 0.86396, 2.18751, 5),
            (0, 0.86396, 1.2, 0),
            (0.07478, 0.87971, 2.18751, 5),
            (0, 0.87971, 1.2, 0),
            (0.07478, 0.89001, 2.18751, 5),
            (0.07478, 0.85, 1.2, 0),
            (0.28333, 0.86396, 2.18751, 5),
            (0.07478, 0.86396, 1.2, 0),
            (0.28333, 0.89001, 2.18751, 5),
            (0.28333, 0.85, 1.2, 0),
            (0.31332, 0.89001, 2.18751, 5),
            (0.31332, 0.85, 1.2, 0),
            (1.2, 0.89001, 2.18751, 5),
            (0, 0.89001, 1.2, 0),
            (0.07478, 0.91727, 1.8668, 5),
            (0, 0.89001, 1.8668, 0),
            (0.07478, 0.91727, 2.18751, 5),
            (0, 0.91727, 1.2, 0),
            (0.07478, 0.93, 2.18751, 5),
            (0.07478, 0.89001, 1.2, 0),
            (0.28333, 0.91727, 2.18751, 5),
            (0.07478, 0.91727, 1.2, 0),
            (0.28333, 0.93, 2.18751, 5),
            (0.28333, 0.89001, 1.2, 0),
            (0.31332, 0.93, 2.18751, 5),
            (0.31332, 0.89001, 1.2, 0),
            (1.2, 0.93, 2.18751, 5),
            (0, 0.85, 2.18751, 0),
            (0.07478, 0.86396, 3.49134, 5),
            (0, 0.85, 3.49134, 0),
            (0.07478, 0.86396, 4, 5),
            (0, 0.86396, 2.18751, 0),
            (0.07478, 0.87971, 4, 5),
            (0, 0.87971, 2.18751, 0),
            (0.07478, 0.89001, 4, 5),
            (0.07478, 0.85, 2.18751, 0),
            (0.28333, 0.86396, 4, 5),
            (0.07478, 0.86396, 2.18751, 0),
            (0.28333, 0.89001, 4, 5),
            (0.28333, 0.85, 2.18751, 0),
            (0.31332, 0.89001, 4, 5),
            (0.31332, 0.85, 2.18751, 0),
            (1.2, 0.89001, 4, 5),
            (0, 0.89001, 2.18751, 0),
            (0.07478, 0.91727, 3.49134, 5),
            (0, 0.89001, 3.49134, 0),
            (0.07478, 0.91727, 4, 5),
            (0, 0.91727, 2.18751, 0),
            (0.07478, 0.93, 3.49134, 5),
            (0, 0.91727, 3.49134, 0),
            (0.07478, 0.93, 4, 5),
            (0.07478, 0.89001, 2.18751, 0),
            (0.28333, 0.91727, 4, 5),
            (0.07478, 0.91727, 2.18751, 0),
            (0.28333, 0.93, 4, 5),
            (0.28333, 0.89001, 2.18751, 0),
            (0.31332, 0.93, 4, 5),
            (0.31332, 0.89001, 2.18751, 0),
            (1.2, 0.93, 4, 5),
        ]
        m.pw = PiecewiseLinearFunction(points=points, function=f)

        # The big win is if the above runs, but we'll check the approximation
        # computationally at least, to make sure that at all the points we gave,
        # the pw linear approximation evaluates to the same value as the
        # original nonlinear function.
        for pt in points:
            self.assertAlmostEqual(m.pw(*pt), f(*pt))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Problem Writer for BARON .bar Format Files
#

import itertools
import logging
import math
from io import StringIO
from contextlib import nullcontext

from pyomo.common.collections import OrderedSet
from pyomo.opt import ProblemFormat
from pyomo.opt.base import AbstractProblemWriter, WriterFactory
from pyomo.core.expr.numvalue import (
    value,
    native_numeric_types,
    native_types,
    nonpyomo_leaf_types,
)
from pyomo.core.expr.visitor import _ToStringVisitor
import pyomo.core.expr as EXPR
from pyomo.core.base import (
    SortComponents,
    SymbolMap,
    ShortNameLabeler,
    NumericLabeler,
    Constraint,
    Objective,
    Param,
)
from pyomo.core.base.component import ActiveComponent

# CLH: EXPORT suffixes "constraint_types" and "branching_priorities"
#     pass their respective information to the .bar file
import pyomo.core.base.suffix
import pyomo.core.kernel.suffix
from pyomo.core.kernel.block import IBlock
from pyomo.repn.util import valid_expr_ctypes_minlp, valid_active_ctypes_minlp, ftoa

logger = logging.getLogger('pyomo.core')


def _handle_PowExpression(visitor, node, values):
    # Per the BARON manual, x ^ y is allowed as long as x and y are not
    # both variables.  There is an issue that if one of the arguments
    # contains "0*var", Pyomo will see that as fixed, but Baron will see
    # it as variable.  We will work around that by resolving any fixed
    # expressions to their corresponding fixed value.
    unfixed_count = 0
    for i, arg in enumerate(node.args):
        if type(arg) in native_types:
            pass
        elif arg.is_fixed():
            values[i] = ftoa(value(arg), True)
        else:
            unfixed_count += 1

    if unfixed_count < 2:
        return f"{values[0]} ^ {values[1]}"
    else:
        return f"exp(({values[0]}) * log({values[1]}))"


_allowableUnaryFunctions = {'exp', 'log10', 'log', 'sqrt'}

_log10_e = ftoa(math.log10(math.e))


def _handle_UnaryFunctionExpression(visitor, node, values):
    if node.name == "sqrt":
        # Parens are necessary because sqrt() and "^" have different
        # precedence levels.  Instead of parsing the arg, be safe and
        # explicitly add parens
        return f"(({values[0]}) ^ 0.5)"
    elif node.name == 'log10':
        return f"({_log10_e} * log({values[0]}))"
    elif node.name not in _allowableUnaryFunctions:
        raise RuntimeError(
            'The BARON .BAR format does not support the unary '
            'function "%s".' % (node.name,)
        )
    return node._to_string(values, visitor.verbose, visitor.smap)


def _handle_AbsExpression(visitor, node, values):
    # Parens are necessary because abs() and "^" have different
    # precedence levels.  Instead of parsing the arg, be safe and
    # explicitly add parens
    return f"((({values[0]}) ^ 2) ^ 0.5)"


_plusMinusOne = {-1, 1}


#
# A visitor pattern that creates a string for an expression
# that is compatible with the BARON syntax.
#
class ToBaronVisitor(_ToStringVisitor):
    _expression_handlers = {
        EXPR.PowExpression: _handle_PowExpression,
        EXPR.UnaryFunctionExpression: _handle_UnaryFunctionExpression,
        EXPR.AbsExpression: _handle_AbsExpression,
    }

    def __init__(self, variables, smap):
        super(ToBaronVisitor, self).__init__(False, smap)
        self.variables = variables

    def visiting_potential_leaf(self, node):
        """
        Visiting a potential leaf.

        Return True if the node is not expanded.
        """
        if node.__class__ in native_types:
            return True, ftoa(node, True)

        if node.is_expression_type():
            # Special handling if NPV and semi-NPV types:
            if not node.is_potentially_variable():
                return True, ftoa(node(), True)
            if node.__class__ is EXPR.MonomialTermExpression:
                return True, self._monomial_to_string(node)
            if node.__class__ is EXPR.LinearExpression:
                return True, self._linear_to_string(node)
            # we will descend into this, so type checking will happen later
            return False, None

        if node.is_component_type():
            if node.ctype not in valid_expr_ctypes_minlp:
                # Make sure all components in active constraints
                # are basic ctypes we know how to deal with.
                raise RuntimeError(
                    "Unallowable component '%s' of type %s found in an active "
                    "constraint or objective.\nThe GAMS writer cannot export "
                    "expressions with this component type."
                    % (node.name, node.ctype.__name__)
                )

        if node.is_fixed():
            return True, ftoa(node(), True)
        else:
            assert node.is_variable_type()
            self.variables.add(id(node))
            return True, self.smap.getSymbol(node)

    def _monomial_to_string(self, node):
        const, var = node.args
        if const.__class__ not in native_types:
            const = value(const)
        if var.is_fixed():
            return ftoa(const * var.value, True)
        # Special handling: ftoa is slow, so bypass _to_string when this
        # is a trivial term
        if not const:
            return '0'
        self.variables.add(id(var))
        if const in _plusMinusOne:
            if const < 0:
                return '-' + self.smap.getSymbol(var)
            else:
                return self.smap.getSymbol(var)
        return ftoa(const, True) + '*' + self.smap.getSymbol(var)

    def _var_to_string(self, node):
        if node.is_fixed():
            return ftoa(node.value, True)
        self.variables.add(id(node))
        return self.smap.getSymbol(node)

    def _linear_to_string(self, node):
        values = [
            (
                self._monomial_to_string(arg)
                if arg.__class__ is EXPR.MonomialTermExpression
                else (
                    ftoa(arg)
                    if arg.__class__ in native_numeric_types
                    else (
                        self._var_to_string(arg)
                        if arg.is_variable_type()
                        else ftoa(value(arg), True)
                    )
                )
            )
            for arg in node.args
        ]
        return node._to_string(values, False, self.smap)


def expression_to_string(expr, variables, smap):
    return ToBaronVisitor(variables, smap).dfs_postorder_stack(expr)


# TODO: The to_string function is handy, but the fact that
#       it calls .name under the hood for all components
#       everywhere they are used will present ENORMOUS
#       overhead for components that have a large index set.
#       It might be worth adding an extra keyword to that
#       function that takes a "labeler" or "symbol_map" for
#       writing non-expression components.


@WriterFactory.register('bar', 'Generate the corresponding BARON BAR file.')
class ProblemWriter_bar(AbstractProblemWriter):
    def __init__(self):
        AbstractProblemWriter.__init__(self, ProblemFormat.bar)

    def _write_equations_section(
        self,
        model,
        output_file,
        all_blocks_list,
        active_components_data_var,
        symbol_map,
        c_labeler,
        output_fixed_variable_bounds,
        skip_trivial_constraints,
        sorter,
    ):
        referenced_variable_ids = OrderedSet()

        def _skip_trivial(constraint_data):
            if skip_trivial_constraints:
                if constraint_data._linear_canonical_form:
                    repn = constraint_data.canonical_form()
                    if (repn.variables is None) or (len(repn.variables) == 0):
                        return True
                elif constraint_data.body.polynomial_degree() == 0:
                    return True
            return False

        #
        # Check for active suffixes to export
        #
        if isinstance(model, IBlock):
            suffix_gen = lambda b: (
                (suf.storage_key, suf)
                for suf in pyomo.core.kernel.suffix.export_suffix_generator(
                    b, active=True, descend_into=False
                )
            )
        else:
            suffix_gen = (
                lambda b: pyomo.core.base.suffix.active_export_suffix_generator(b)
            )
        r_o_eqns = {}
        c_eqns = {}
        l_eqns = {}
        branching_priorities_suffixes = []
        for block in all_blocks_list:
            for name, suffix in suffix_gen(block):
                if name in {'branching_priorities', 'priority'}:
                    branching_priorities_suffixes.append(suffix)
                elif name == 'constraint_types':
                    for constraint_data, constraint_type in suffix.items():
                        info = constraint_data.to_bounded_expression(True)
                        if not _skip_trivial(constraint_data):
                            if constraint_type.lower() == 'relaxationonly':
                                r_o_eqns[constraint_data] = info
                            elif constraint_type.lower() == 'convex':
                                c_eqns[constraint_data] = info
                            elif constraint_type.lower() == 'local':
                                l_eqns[constraint_data] = info
                            else:
                                raise ValueError(
                                    "A suffix '%s' contained an invalid value: %s\n"
                                    "Choices are: [relaxationonly, convex, local]"
                                    % (suffix.name, constraint_type)
                                )
                else:
                    if block is block.model():
                        if block.name == 'unknown':
                            _location = 'model'
                        else:
                            _location = "model '%s'" % (block.name,)
                    else:
                        _location = "block '%s'" % (block.name,)

                    raise ValueError(
                        "The BARON writer can not export suffix with name '%s'. "
                        "Either remove it from the %s or deactivate it."
                        % (name, _location)
                    )

        non_standard_eqns = set()
        non_standard_eqns.update(r_o_eqns)
        non_standard_eqns.update(c_eqns)
        non_standard_eqns.update(l_eqns)

        #
        # EQUATIONS
        #

        # Equation Declaration
        n_roeqns = len(r_o_eqns)
        n_ceqns = len(c_eqns)
        n_leqns = len(l_eqns)
        eqns = {}

        # Alias the constraints by declaration order since Baron does not
        # include the constraint names in the solution file. It is important
        # that this alias not clash with any real constraint labels, hence
        # the use of the ".c<integer>" template. It is not possible to declare
        # a component having this type of name when using standard syntax.
        # There are ways to do it, but it is unlikely someone will.
        order_counter = 0
        alias_template = ".c%d"
        output_file.write('EQUATIONS ')
        output_file.write("c_e_FIX_ONE_VAR_CONST__")
        order_counter += 1
        for block in all_blocks_list:
            for constraint_data in block.component_data_objects(
                Constraint, active=True, sort=sorter, descend_into=False
            ):
                lb, body, ub = constraint_data.to_bounded_expression(True)
                if lb is None and ub is None:
                    assert not constraint_data.equality
                    continue  # non-binding, so skip

                if (not _skip_trivial(constraint_data)) and (
                    constraint_data not in non_standard_eqns
                ):
                    eqns[constraint_data] = lb, body, ub

                    con_symbol = symbol_map.createSymbol(constraint_data, c_labeler)
                    assert not con_symbol.startswith('.')
                    assert con_symbol != "c_e_FIX_ONE_VAR_CONST__"

                    symbol_map.alias(constraint_data, alias_template % order_counter)
                    output_file.write(", " + str(con_symbol))
                    order_counter += 1

        output_file.write(";\n\n")

        if n_roeqns > 0:
            output_file.write('RELAXATION_ONLY_EQUATIONS ')
            for i, constraint_data in enumerate(r_o_eqns):
                con_symbol = symbol_map.createSymbol(constraint_data, c_labeler)
                assert not con_symbol.startswith('.')
                assert con_symbol != "c_e_FIX_ONE_VAR_CONST__"
                symbol_map.alias(constraint_data, alias_template % order_counter)
                if i == n_roeqns - 1:
                    output_file.write(str(con_symbol) + ';\n\n')
                else:
                    output_file.write(str(con_symbol) + ', ')
                order_counter += 1

        if n_ceqns > 0:
            output_file.write('CONVEX_EQUATIONS ')
            for i, constraint_data in enumerate(c_eqns):
                con_symbol = symbol_map.createSymbol(constraint_data, c_labeler)
                assert not con_symbol.startswith('.')
                assert con_symbol != "c_e_FIX_ONE_VAR_CONST__"
                symbol_map.alias(constraint_data, alias_template % order_counter)
                if i == n_ceqns - 1:
                    output_file.write(str(con_symbol) + ';\n\n')
                else:
                    output_file.write(str(con_symbol) + ', ')
                order_counter += 1

        if n_leqns > 0:
            output_file.write('LOCAL_EQUATIONS ')
            for i, constraint_data in enumerate(l_eqns):
                con_symbol = symbol_map.createSymbol(constraint_data, c_labeler)
                assert not con_symbol.startswith('.')
                assert con_symbol != "c_e_FIX_ONE_VAR_CONST__"
                symbol_map.alias(constraint_data, alias_template % order_counter)
                if i == n_leqns - 1:
                    output_file.write(str(con_symbol) + ';\n\n')
                else:
                    output_file.write(str(con_symbol) + ', ')
                order_counter += 1

        # Create a dictionary of baron variable names to match to the
        # strings that constraint.to_string() prints. An important
        # note is that the variable strings are padded by spaces so
        # that whole variable names are recognized, and simple
        # variable names are not identified inside longer names.
        # Example: ' x[1] ' -> ' x3 '
        # FIXME: 7/18/14 CLH: This may cause mistakes if spaces in
        #                    variable names are allowed
        if isinstance(model, IBlock):
            mutable_param_gen = lambda b: b.components(ctype=Param, descend_into=False)
        else:

            def mutable_param_gen(b):
                for param in block.component_objects(Param):
                    if param.mutable and param.is_indexed():
                        param_data_iter = (
                            param_data for index, param_data in param.items()
                        )
                    elif not param.is_indexed():
                        param_data_iter = iter([param])
                    else:
                        param_data_iter = iter([])

                    for param_data in param_data_iter:
                        yield param_data

        # Equation Definition
        output_file.write('c_e_FIX_ONE_VAR_CONST__:  ONE_VAR_CONST__  == 1;\n')
        for constraint_data, (lb, body, ub) in itertools.chain(
            eqns.items(), r_o_eqns.items(), c_eqns.items(), l_eqns.items()
        ):
            variables = OrderedSet()
            # print(symbol_map.byObject.keys())
            eqn_body = expression_to_string(body, variables, smap=symbol_map)
            # print(symbol_map.byObject.keys())
            referenced_variable_ids.update(variables)

            if len(variables) == 0:
                assert not skip_trivial_constraints
                eqn_body += " + 0 * ONE_VAR_CONST__ "

            # 7/29/14 CLH:
            # FIXME: Baron doesn't handle many of the
            #       intrinsic_functions available in pyomo. The
            #       error message given by baron is also very
            #       weak.  Either a function here to re-write
            #       unallowed expressions or a way to track solver
            #       capability by intrinsic_expression would be
            #       useful.
            ##########################

            con_symbol = symbol_map.byObject[id(constraint_data)]
            output_file.write(str(con_symbol) + ': ')

            # Fill in the left and right hand side (constants) of
            #  the equations

            # Equality constraint
            if constraint_data.equality:
                eqn_lhs = ''
                eqn_rhs = ' == ' + ftoa(ub)

            # Greater than constraint
            elif ub is None:
                eqn_rhs = ' >= ' + ftoa(lb)
                eqn_lhs = ''

            # Less than constraint
            elif lb is None:
                eqn_rhs = ' <= ' + ftoa(ub)
                eqn_lhs = ''

            # Double-sided constraint
            elif lb is not None and ub is not None:
                eqn_lhs = ftoa(lb) + ' <= '
                eqn_rhs = ' <= ' + ftoa(ub)

            eqn_string = eqn_lhs + eqn_body + eqn_rhs + ';\n'
            output_file.write(eqn_string)

        #
        # OBJECTIVE
        #

        output_file.write("\nOBJ: ")

        n_objs = 0
        for block in all_blocks_list:
            for objective_data in block.component_data_objects(
                Objective, active=True, sort=sorter, descend_into=False
            ):
                n_objs += 1
                if n_objs > 1:
                    raise ValueError(
                        "The BARON writer has detected multiple active "
                        "objective functions on model %s, but "
                        "currently only handles a single objective." % (model.name)
                    )

                # create symbol
                symbol_map.createSymbol(objective_data, c_labeler)
                symbol_map.alias(objective_data, "__default_objective__")

                if objective_data.is_minimizing():
                    output_file.write("minimize ")
                else:
                    output_file.write("maximize ")

                variables = OrderedSet()
                # print(symbol_map.byObject.keys())
                obj_string = expression_to_string(
                    objective_data.expr, variables, smap=symbol_map
                )
                # print(symbol_map.byObject.keys())
                referenced_variable_ids.update(variables)

        output_file.write(obj_string + ";\n\n")
        # referenced_variable_ids.update(symbol_map.byObject.keys())

        return referenced_variable_ids, branching_priorities_suffixes

    def __call__(self, model, output_filename, solver_capability, io_options):
        if output_filename is None:
            output_filename = model.name + ".bar"

        # If the user provides a file name, we will use the opened file
        # as a context manager to ensure it gets closed.  If the user
        # provided something else (e.g., a file-like object), we will
        # use a nullcontext manager to prevent closing it.
        if isinstance(output_filename, str):
            output_file = open(output_filename, "w")
        else:
            output_file = nullcontext(output_filename)

        with output_file as FILE:
            symbol_map = self._write_bar_file(
                model, FILE, solver_capability, io_options
            )

        return output_filename, symbol_map

    def _write_bar_file(self, model, output_file, solver_capability, io_options):
        # Make sure not to modify the user's dictionary, they may be
        # reusing it outside of this call
        io_options = dict(io_options)

        # NOTE: io_options is a simple dictionary of keyword-value
        #       pairs specific to this writer.
        symbolic_solver_labels = io_options.pop("symbolic_solver_labels", False)
        labeler = io_options.pop("labeler", None)

        # How much effort do we want to put into ensuring the
        # LP file is written deterministically for a Pyomo model:
        #    0 : None
        #    1 : sort keys of indexed components (default)
        #    2 : sort keys AND sort names (over declaration order)
        file_determinism = io_options.pop("file_determinism", 1)

        sorter = SortComponents.unsorted
        if file_determinism >= 1:
            sorter = sorter | SortComponents.indices
            if file_determinism >= 2:
                sorter = sorter | SortComponents.alphabetical

        output_fixed_variable_bounds = io_options.pop(
            "output_fixed_variable_bounds", False
        )

        # Skip writing constraints whose body section is fixed (i.e.,
        # no variables)
        skip_trivial_constraints = io_options.pop("skip_trivial_constraints", False)

        # Note: Baron does not allow specification of runtime
        #       option outside of this file, so we add support
        #       for them here
        solver_options = io_options.pop("solver_options", {})

        if len(io_options):
            raise ValueError(
                "ProblemWriter_baron_writer passed unrecognized io_options:\n\t"
                + "\n\t".join("%s = %s" % (k, v) for k, v in io_options.items())
            )

        if symbolic_solver_labels and (labeler is not None):
            raise ValueError(
                "Baron problem writer: Using both the "
                "'symbolic_solver_labels' and 'labeler' "
                "I/O options is forbidden"
            )

        # Make sure there are no strange ActiveComponents. The expression
        # walker will handle strange things in constraints later.
        model_ctypes = model.collect_ctypes(active=True)
        invalids = set()
        for t in model_ctypes - valid_active_ctypes_minlp:
            if issubclass(t, ActiveComponent):
                invalids.add(t)
        if len(invalids):
            invalids = [t.__name__ for t in invalids]
            raise RuntimeError(
                "Unallowable active component(s) %s.\nThe BARON writer cannot "
                "export models with this component type." % ", ".join(invalids)
            )

        # Process the options. Rely on baron to catch
        # and reset bad option values
        output_file.write("OPTIONS {\n")
        summary_found = False
        if len(solver_options):
            for key, val in solver_options.items():
                if key.lower() == 'summary':
                    summary_found = True
                if key.endswith("Name"):
                    output_file.write(key + ": \"" + str(val) + "\";\n")
                else:
                    output_file.write(key + ": " + str(val) + ";\n")
        if not summary_found:
            # The 'summary option is defaulted to 0, so that no
            # summary file is generated in the directory where the
            # user calls baron. Check if a user explicitly asked for
            # a summary file.
            output_file.write("Summary: 0;\n")
        output_file.write("}\n\n")

        if symbolic_solver_labels:
            # Note that the Var and Constraint labelers must use the
            # same labeler, so that we can correctly detect name
            # collisions (which can arise when we truncate the labels to
            # the max allowable length.  BARON requires all identifiers
            # to start with a letter.  We will (randomly) choose "s_"
            # (for 'shortened')
            v_labeler = c_labeler = ShortNameLabeler(
                15,
                prefix='s_',
                suffix='_',
                caseInsensitive=True,
                legalRegex='^[a-zA-Z]',
            )
        elif labeler is None:
            v_labeler = NumericLabeler('x')
            c_labeler = NumericLabeler('c')
        else:
            v_labeler = c_labeler = labeler

        symbol_map = SymbolMap()
        symbol_map.default_labeler = v_labeler
        # sm_bySymbol = symbol_map.bySymbol

        # Cache the list of model blocks so we don't have to call
        # model.block_data_objects() many many times, which is slow
        # for indexed blocks
        all_blocks_list = list(
            model.block_data_objects(active=True, sort=sorter, descend_into=True)
        )
        active_components_data_var = {}
        # for block in all_blocks_list:
        #    tmp = active_components_data_var[id(block)] = \
        #          list(obj for obj in block.component_data_objects(Var,
        #                                                           sort=sorter,
        #                                                           descend_into=False))
        #    create_symbols_func(symbol_map, tmp, labeler)

        # GAH: Not sure this is necessary, and also it would break for
        #      non-mutable indexed params so I am commenting out for now.
        # for param_data in active_components_data(block, Param, sort=sorter):
        # instead of checking if param_data.mutable:
        # if not param_data.is_constant():
        #    create_symbol_func(symbol_map, param_data, labeler)

        # symbol_map_variable_ids = set(symbol_map.byObject.keys())
        # object_symbol_dictionary = symbol_map.byObject

        #
        # Go through the objectives and constraints and generate
        # the output so that we can obtain the set of referenced
        # variables.
        #
        equation_section_stream = StringIO()
        referenced_variable_ids, branching_priorities_suffixes = (
            self._write_equations_section(
                model,
                equation_section_stream,
                all_blocks_list,
                active_components_data_var,
                symbol_map,
                c_labeler,
                output_fixed_variable_bounds,
                skip_trivial_constraints,
                sorter,
            )
        )

        #
        # BINARY_VARIABLES, INTEGER_VARIABLES, POSITIVE_VARIABLES, VARIABLES
        #

        BinVars = []
        IntVars = []
        PosVars = []
        Vars = []
        for vid in referenced_variable_ids:
            name = symbol_map.byObject[vid]
            var_data = symbol_map.bySymbol[name]

            if var_data.is_continuous():
                if var_data.has_lb() and (value(var_data.lb) >= 0):
                    TypeList = PosVars
                else:
                    TypeList = Vars
            elif var_data.is_binary():
                TypeList = BinVars
            elif var_data.is_integer():
                TypeList = IntVars
            else:
                assert False
            TypeList.append(name)

        if len(BinVars) > 0:
            BinVars.sort()
            output_file.write('BINARY_VARIABLES ')
            output_file.write(", ".join(BinVars))
            output_file.write(';\n\n')

        if len(IntVars) > 0:
            IntVars.sort()
            output_file.write('INTEGER_VARIABLES ')
            output_file.write(", ".join(IntVars))
            output_file.write(';\n\n')

        PosVars.append('ONE_VAR_CONST__')
        PosVars.sort()
        output_file.write('POSITIVE_VARIABLES ')
        output_file.write(", ".join(PosVars))
        output_file.write(';\n\n')

        if len(Vars) > 0:
            Vars.sort()
            output_file.write('VARIABLES ')
            output_file.write(", ".join(Vars))
            output_file.write(';\n\n')

        #
        # LOWER_BOUNDS
        #

        lbounds = {}
        for vid in referenced_variable_ids:
            name = symbol_map.byObject[vid]
            var_data = symbol_map.bySymbol[name]

            if var_data.fixed:
                if output_fixed_variable_bounds:
                    var_data_lb = ftoa(var_data.value, False)
                else:
                    var_data_lb = None
            else:
                var_data_lb = None
                if var_data.has_lb():
                    var_data_lb = ftoa(var_data.lb, False)

            if var_data_lb is not None:
                name_to_output = symbol_map.getSymbol(var_data)
                lbounds[name_to_output] = '%s: %s;\n' % (name_to_output, var_data_lb)

        if len(lbounds) > 0:
            output_file.write("LOWER_BOUNDS{\n")
            output_file.write("".join(lbounds[key] for key in sorted(lbounds.keys())))
            output_file.write("}\n\n")
        lbounds = None

        #
        # UPPER_BOUNDS
        #

        ubounds = {}
        for vid in referenced_variable_ids:
            name = symbol_map.byObject[vid]
            var_data = symbol_map.bySymbol[name]

            if var_data.fixed:
                if output_fixed_variable_bounds:
                    var_data_ub = ftoa(var_data.value, False)
                else:
                    var_data_ub = None
            else:
                var_data_ub = None
                if var_data.has_ub():
                    var_data_ub = ftoa(var_data.ub, False)

            if var_data_ub is not None:
                name_to_output = symbol_map.getSymbol(var_data)
                ubounds[name_to_output] = '%s: %s;\n' % (name_to_output, var_data_ub)

        if len(ubounds) > 0:
            output_file.write("UPPER_BOUNDS{\n")
            output_file.write("".join(ubounds[key] for key in sorted(ubounds.keys())))
            output_file.write("}\n\n")
        ubounds = None

        #
        # BRANCHING_PRIORITIES
        #

        # Specifying priorities requires that the pyomo model has established an
        # EXTERNAL, float suffix called 'branching_priorities' on the model
        # object, indexed by the relevant variable
        BranchingPriorityHeader = False
        for suffix in branching_priorities_suffixes:
            for var, priority in suffix.items():
                if var.is_indexed():
                    var_iter = var.values()
                else:
                    var_iter = (var,)
                for var_data in var_iter:
                    if id(var_data) not in referenced_variable_ids:
                        continue
                    if priority is not None:
                        if not BranchingPriorityHeader:
                            output_file.write('BRANCHING_PRIORITIES{\n')
                            BranchingPriorityHeader = True
                        output_file.write(
                            "%s: %s;\n" % (symbol_map.getSymbol(var_data), priority)
                        )

        if BranchingPriorityHeader:
            output_file.write("}\n\n")

        #
        # Now write the objective and equations section
        #
        output_file.write(equation_section_stream.getvalue())

        #
        # STARTING_POINT
        #
        output_file.write('STARTING_POINT{\nONE_VAR_CONST__: 1;\n')
        tmp = {}
        for vid in referenced_variable_ids:
            name = symbol_map.byObject[vid]
            var_data = symbol_map.bySymbol[name]

            starting_point = var_data.value
            if starting_point is not None:
                var_name = symbol_map.getSymbol(var_data)
                tmp[var_name] = "%s: %s;\n" % (var_name, ftoa(starting_point, False))

        output_file.write("".join(tmp[key] for key in sorted(tmp.keys())))
        output_file.write('}\n\n')

        return symbol_map



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import pandas as pd
from os.path import join, abspath, dirname

import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.examples.reactor_design.reactor_design import (
    ReactorDesignExperiment,
)


class TimeSeriesReactorDesignExperiment(ReactorDesignExperiment):

    def __init__(self, data, experiment_number):
        self.data = data
        self.experiment_number = experiment_number
        data_i = data.loc[data['experiment'] == experiment_number, :]
        self.data_i = data_i.reset_index()
        self.model = None

    def finalize_model(self):
        m = self.model

        # Experiment inputs values
        m.sv = self.data_i['sv'].mean()
        m.caf = self.data_i['caf'].mean()

        # Experiment output values
        m.ca = self.data_i['ca'][0]
        m.cb = self.data_i['cb'][0]
        m.cc = self.data_i['cc'][0]
        m.cd = self.data_i['cd'][0]

        return m


def main():
    # Parameter estimation using timeseries data, grouped by experiment number

    # Data, includes multiple sensors for ca and cc
    file_dirname = dirname(abspath(str(__file__)))
    file_name = abspath(join(file_dirname, 'reactor_data_timeseries.csv'))
    data = pd.read_csv(file_name)

    # Create an experiment list
    exp_list = []
    for i in data['experiment'].unique():
        exp_list.append(TimeSeriesReactorDesignExperiment(data, i))

    def SSE_timeseries(model):

        expr = 0
        for y, y_hat in model.experiment_outputs.items():
            num_time_points = len(y_hat)
            for i in range(num_time_points):
                expr += ((y - y_hat[i]) ** 2) * (1 / num_time_points)

        return expr

    # View one model & SSE
    # exp0_model = exp_list[0].get_labeled_model()
    # exp0_model.pprint()
    # print(SSE_timeseries(exp0_model))

    pest = parmest.Estimator(exp_list, obj_function=SSE_timeseries)
    obj, theta = pest.theta_est()
    print(obj)
    print(theta)


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import itertools
import pyomo.common.unittest as unittest
import pyomo.environ as pyo

from pyomo.common.dependencies import networkx_available as nx_available
from pyomo.contrib.pynumero.dependencies import (
    numpy as np,
    numpy_available,
    scipy,
    scipy_available,
)

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pynumero needs scipy and numpy to run NLP tests")

from pyomo.common.dependencies.scipy import sparse as sps

from pyomo.contrib.pynumero.asl import AmplInterface

if not AmplInterface.available():
    raise unittest.SkipTest("Pynumero needs the ASL extension to run cyipopt tests")

from pyomo.contrib.pynumero.algorithms.solvers.cyipopt_solver import cyipopt_available
from pyomo.contrib.pynumero.interfaces.external_pyomo_model import (
    ExternalPyomoModel,
    get_hessian_of_constraint,
)
from pyomo.contrib.pynumero.interfaces.pyomo_grey_box_nlp import (
    PyomoNLPWithGreyBoxBlocks,
)
from pyomo.contrib.pynumero.interfaces.external_grey_box import ExternalGreyBoxBlock
from pyomo.contrib.pynumero.algorithms.solvers.cyipopt_solver import CyIpoptSolver
from pyomo.contrib.pynumero.interfaces.cyipopt_interface import CyIpoptNLP

if not pyo.SolverFactory("ipopt").available():
    raise unittest.SkipTest("Need IPOPT to run ExternalPyomoModel tests")


class SimpleModel1:
    def make_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=2.0)
        m.y = pyo.Var(initialize=2.0)
        m.residual_eqn = pyo.Constraint(expr=m.x**2 + m.y**2 == 1.0)
        m.external_eqn = pyo.Constraint(expr=m.x * m.y == 0.2)
        # The "external function constraint" exposed by the ExternalPyomoModel
        # will look like: x**2 + 0.04/x**2 - 1 == 0
        return m

    def evaluate_external_variables(self, x):
        # y(x)
        return 0.2 / x

    def evaluate_external_jacobian(self, x):
        # dydx
        return -0.2 / (x**2)

    def evaluate_external_hessian(self, x):
        # d2ydx2
        return 0.4 / (x**3)

    def evaluate_residual(self, x):
        return x**2 + 0.04 / x**2 - 1

    def evaluate_jacobian(self, x):
        return 2 * x - 0.08 / x**3

    def evaluate_hessian(self, x):
        return 2 + 0.24 / x**4


class SimpleModel2:
    """
    The purpose of this model is to exercise each term in the computation
    of the d2ydx2 Hessian.
    """

    def make_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=2.0)
        m.y = pyo.Var(initialize=2.0)
        m.residual_eqn = pyo.Constraint(expr=m.x**2 + m.y**2 == 1.0)
        m.external_eqn = pyo.Constraint(expr=(m.x**3) * (m.y**3) == 0.2)
        return m

    def evaluate_external_variables(self, x):
        return 0.2 ** (1 / 3) / x

    def evaluate_external_jacobian(self, x):
        return -(0.2 ** (1 / 3)) / (x**2)

    def evaluate_external_hessian(self, x):
        return 2 * 0.2 ** (1 / 3) / (x**3)

    def evaluate_residual(self, x):
        return x**2 + 0.2 ** (2 / 3) / x**2 - 1

    def evaluate_jacobian(self, x):
        return 2 * x - 2 * 0.2 ** (2 / 3) / x**3

    def evaluate_hessian(self, x):
        return 2 + 6 * 0.2 ** (2 / 3) / x**4


class SimpleModel2by2_1:
    """
    The purpose of this model is to test second derivative computation
    when the external model is nonlinear only in x. This exercises
    the first term in the second derivative implicit function theorem.
    """

    def make_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([0, 1], initialize=2.0)
        m.y = pyo.Var([0, 1], initialize=2.0)

        def residual_eqn_rule(m, i):
            # These equations are chosen to exercise every term in the
            # equality hessian calculation, i.e. to have nonlinearities
            # in x, y, and xy.
            if i == 0:
                return m.x[0] ** 2 + m.x[0] * m.y[0] + m.y[0] ** 2 == 1.0
            elif i == 1:
                return m.x[1] ** 2 + m.x[1] * m.y[1] == 2.0

        m.residual_eqn = pyo.Constraint([0, 1], rule=residual_eqn_rule)

        def external_eqn_rule(m, i):
            if i == 0:
                return m.y[0] + m.y[1] + m.x[0] * m.x[1] + m.x[0] ** 2 == 1.0
            elif i == 1:
                return m.y[0] + 2.0 * m.x[0] * m.x[1] + m.x[1] ** 2 == 2.0

        m.external_eqn = pyo.Constraint([0, 1], rule=external_eqn_rule)

        return m

    # The following three methods are evaluation and derivatives of
    # the equality constraints exposed by the ExternalPyomoModel
    def evaluate_residual(self, x):
        f0 = (
            x[0] ** 2
            + 2 * x[0]
            - 2 * x[0] ** 2 * x[1]
            - x[1] ** 2 * x[0]
            + 4
            - 8 * x[0] * x[1]
            - 4 * x[1] ** 2
            + 4 * x[0] ** 2 * x[1] ** 2
            + 4 * x[0] * x[1] ** 3
            + x[1] ** 4
            - 1.0
        )
        f1 = x[1] ** 2 - x[1] + x[0] * x[1] ** 2 + x[1] ** 3 - x[0] ** 2 * x[1] - 2.0
        return (f0, f1)

    def evaluate_jacobian(self, x):
        df0dx0 = (
            2 * x[0]
            + 2
            - 4 * x[0] * x[1]
            - x[1] ** 2
            - 8 * x[1]
            + 8 * x[0] * x[1] ** 2
            + 4 * x[1] ** 3
        )
        df0dx1 = (
            -2 * x[0] ** 2
            - 2 * x[0] * x[1]
            - 8 * x[0]
            - 8 * x[1]
            + 8 * x[0] ** 2 * x[1]
            + 12 * x[0] * x[1] ** 2
            + 4 * x[1] ** 3
        )
        df1dx0 = x[1] ** 2 - 2 * x[0] * x[1]
        df1dx1 = 2 * x[1] - 1 + 2 * x[0] * x[1] - x[0] ** 2 + 3 * x[1] ** 2
        return np.array([[df0dx0, df0dx1], [df1dx0, df1dx1]])

    def evaluate_hessian(self, x):
        df0dx0dx0 = 2 - 4 * x[1] + 8 * x[1] ** 2
        df0dx0dx1 = -4 * x[0] - 2 * x[1] - 8 + 16 * x[0] * x[1] + 12 * x[1] ** 2
        df0dx1dx1 = -2 * x[0] - 8 + 8 * x[0] ** 2 + 24 * x[0] * x[1] + 12 * x[1] ** 2

        df1dx0dx0 = -2 * x[1]
        df1dx0dx1 = 2 * x[1] - 2 * x[0]
        df1dx1dx1 = 2 + 2 * x[0] + 6 * x[1]
        d2f0 = np.array([[df0dx0dx0, df0dx0dx1], [df0dx0dx1, df0dx1dx1]])
        d2f1 = np.array([[df1dx0dx0, df1dx0dx1], [df1dx0dx1, df1dx1dx1]])
        return [d2f0, d2f1]

    # The following three methods are evaluation and derivatives of
    # the external function "hidden by" the ExternalPyomoModel
    def evaluate_external_variables(self, x):
        y0 = 2.0 - 2.0 * x[0] * x[1] - x[1] ** 2
        y1 = 1.0 - y0 - x[0] * x[1] - x[0] ** 2
        return (y0, y1)

    def evaluate_external_jacobian(self, x):
        dy0dx0 = -2.0 * x[1]
        dy0dx1 = -2.0 * x[0] - 2.0 * x[1]
        dy1dx0 = -dy0dx0 - x[1] - 2.0 * x[0]
        dy1dx1 = -dy0dx1 - x[0]
        return np.array([[dy0dx0, dy0dx1], [dy1dx0, dy1dx1]])

    def evaluate_external_hessian(self, x):
        dy0dx0dx0 = 0.0
        dy0dx0dx1 = -2.0
        dy0dx1dx1 = -2.0

        dy1dx0dx0 = -dy0dx0dx0 - 2.0
        dy1dx0dx1 = -dy0dx0dx1 - 1.0
        dy1dx1dx1 = -dy0dx1dx1

        dy0dxdx = np.array([[dy0dx0dx0, dy0dx0dx1], [dy0dx0dx1, dy0dx1dx1]])
        dy1dxdx = np.array([[dy1dx0dx0, dy1dx0dx1], [dy1dx0dx1, dy1dx1dx1]])
        return [dy0dxdx, dy1dxdx]


class Model2by2:
    """
    The purpose of this model is to test d2ydx2 Hessian computation when
    transposes result in a nontrivial modification of Hessian/Jacobian
    matrices.
    """

    def make_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([0, 1], initialize=2.0)
        m.y = pyo.Var([0, 1], initialize=2.0)

        m.residual_eqn = pyo.Constraint(
            expr=m.x[0] ** 2 + m.x[1] ** 2 + m.y[0] ** 2 + m.y[1] ** 2 == 1.0
        )

        def external_eqn_rule(m, i):
            if i == 0:
                return (m.x[0] ** 2) * m.y[0] * (m.x[1] ** 0.5) * m.y[1] - 0.1 == 0
            elif i == 1:
                return m.x[0] * m.y[0] * m.x[1] - 0.2 == 0

        m.external_eqn = pyo.Constraint([0, 1], rule=external_eqn_rule)

        return m

    def evaluate_external_variables(self, x):
        y0 = 0.2 / (x[0] * x[1])
        y1 = 0.1 / (x[0] ** 2 * y0 * x[1] ** 0.5)
        return [y0, y1]

    def evaluate_external_jacobian(self, x):
        dy0dx0 = -0.2 / (x[0] ** 2 * x[1])
        dy0dx1 = -0.2 / (x[0] * x[1] ** 2)
        dy1dx0 = -0.5 * x[1] ** 0.5 / x[0] ** 2
        dy1dx1 = 0.25 / (x[0] * x[1] ** 0.5)
        return np.array([[dy0dx0, dy0dx1], [dy1dx0, dy1dx1]])

    def evaluate_external_hessian(self, x):
        d2y0dx0dx0 = 0.4 / (x[0] ** 3 * x[1])
        d2y0dx0dx1 = 0.2 / (x[0] ** 2 * x[1] ** 2)
        d2y0dx1dx1 = 0.4 / (x[0] * x[1] ** 3)

        d2y1dx0dx0 = x[1] ** 0.5 / (x[0] ** 3)
        d2y1dx0dx1 = -0.25 / (x[0] ** 2 * x[1] ** 0.5)
        d2y1dx1dx1 = -0.125 / (x[0] * x[1] ** 1.5)

        d2y0dxdx = np.array([[d2y0dx0dx0, d2y0dx0dx1], [d2y0dx0dx1, d2y0dx1dx1]])
        d2y1dxdx = np.array([[d2y1dx0dx0, d2y1dx0dx1], [d2y1dx0dx1, d2y1dx1dx1]])
        return [d2y0dxdx, d2y1dxdx]

    def calculate_external_multipliers(self, lam, x):
        r"""
        Calculates the multipliers of the external constraints
        from the multipliers of the residual constraints,
        assuming zero dual infeasibility in the coordinates of
        the external variables.
        This is calculated analytically from:

        \nabla_y f^T \lambda_f + \nabla_y g^T \lambda_g = 0

        """
        y = self.evaluate_external_variables(x)
        lg0 = -2 * y[1] * lam[0] / (x[0] ** 2 * x[1] ** 0.5 * y[0])
        lg1 = -(2 * y[0] * lam[0] + x[0] ** 2 * x[1] ** 0.5 * y[1] * lg0) / (
            x[0] * x[1]
        )
        return [lg0, lg1]

    def calculate_full_space_lagrangian_hessians(self, lam, x):
        y = self.evaluate_external_variables(x)
        lam_g = self.calculate_external_multipliers(lam, x)
        d2fdx0dx0 = 2.0
        d2fdx1dx1 = 2.0
        d2fdy0dy0 = 2.0
        d2fdy1dy1 = 2.0
        hfxx = np.array([[d2fdx0dx0, 0], [0, d2fdx1dx1]])
        hfxy = np.array([[0, 0], [0, 0]])
        hfyy = np.array([[d2fdy0dy0, 0], [0, d2fdy1dy1]])

        dg0dx0dx0 = 2 * y[0] * x[1] ** 0.5 * y[1]
        dg0dx0dx1 = x[0] * y[0] * y[1] / x[1] ** 0.5
        dg0dx1dx1 = -1 / 4 * x[0] ** 2 * y[0] * y[1] / x[1] ** (3 / 2)
        dg0dx0dy0 = 2 * x[0] * x[1] ** 0.5 * y[1]
        dg0dx0dy1 = 2 * x[0] * y[0] * x[1] ** 0.5
        dg0dx1dy0 = 0.5 * x[0] ** 2 * y[1] / x[1] ** 0.5
        dg0dx1dy1 = 0.5 * x[0] ** 2 * y[0] / x[1] ** 0.5
        dg0dy0dy1 = x[0] ** 2 * x[1] ** 0.5
        hg0xx = np.array([[dg0dx0dx0, dg0dx0dx1], [dg0dx0dx1, dg0dx1dx1]])
        hg0xy = np.array([[dg0dx0dy0, dg0dx0dy1], [dg0dx1dy0, dg0dx1dy1]])
        hg0yy = np.array([[0, dg0dy0dy1], [dg0dy0dy1, 0]])

        dg1dx0dx1 = y[0]
        dg1dx0dy0 = x[1]
        dg1dx1dy0 = x[0]
        hg1xx = np.array([[0, dg1dx0dx1], [dg1dx0dx1, 0]])
        hg1xy = np.array([[dg1dx0dy0, 0], [dg1dx1dy0, 0]])
        hg1yy = np.zeros((2, 2))

        hlxx = lam[0] * hfxx + lam_g[0] * hg0xx + lam_g[1] * hg1xx
        hlxy = lam[0] * hfxy + lam_g[0] * hg0xy + lam_g[1] * hg1xy
        hlyy = lam[0] * hfyy + lam_g[0] * hg0yy + lam_g[1] * hg1yy
        return hlxx, hlxy, hlyy

    def calculate_reduced_lagrangian_hessian(self, lam, x):
        dydx = self.evaluate_external_jacobian(x)
        hlxx, hlxy, hlyy = self.calculate_full_space_lagrangian_hessians(lam, x)
        return (
            hlxx
            + (hlxy.dot(dydx)).transpose()
            + hlxy.dot(dydx)
            + dydx.transpose().dot(hlyy).dot(dydx)
        )


class TestGetHessianOfConstraint(unittest.TestCase):
    def test_simple_model_1(self):
        model = SimpleModel1()
        m = model.make_model()
        m.x.set_value(2.0)
        m.y.set_value(2.0)

        con = m.residual_eqn
        expected_hess = np.array([[2.0, 0.0], [0.0, 2.0]])
        hess = get_hessian_of_constraint(con)
        self.assertTrue(np.all(expected_hess == hess.toarray()))

        expected_hess = np.array([[2.0]])
        hess = get_hessian_of_constraint(con, [m.x])
        self.assertTrue(np.all(expected_hess == hess.toarray()))

        con = m.external_eqn
        expected_hess = np.array([[0.0, 1.0], [1.0, 0.0]])
        hess = get_hessian_of_constraint(con)
        self.assertTrue(np.all(expected_hess == hess.toarray()))

    def test_polynomial(self):
        m = pyo.ConcreteModel()

        n_x = 3
        x1 = 1.1
        x2 = 1.2
        x3 = 1.3
        m.x = pyo.Var(range(1, n_x + 1), initialize={1: x1, 2: x2, 3: x3})
        m.eqn = pyo.Constraint(
            expr=5 * (m.x[1] ** 5)  # T1
            + 5 * (m.x[1] ** 4) * (m.x[2])  # T2
            + 5 * (m.x[1] ** 3) * (m.x[2]) * (m.x[3])  # T3
            + 5 * (m.x[1]) * (m.x[2] ** 2) * (m.x[3] ** 2)  # T4
            + 4 * (m.x[1] ** 2) * (m.x[2]) * (m.x[3])  # T5
            + 4 * (m.x[2] ** 2) * (m.x[3] ** 2)  # T6
            + 4 * (m.x[3] ** 4)  # T7
            + 3 * (m.x[1]) * (m.x[2]) * (m.x[3])  # T8
            + 3 * (m.x[2] ** 3)  # T9
            + 3 * (m.x[2] ** 2) * (m.x[3])  # T10
            + 2 * (m.x[1]) * (m.x[2])  # T11
            + 2 * (m.x[2]) * (m.x[3])  # T12
            == 0
        )

        rcd = []
        rcd.append(
            (
                0,
                0,
                (
                    # wrt x1, x1
                    5 * 5 * 4 * x1**3  # T1
                    + 5 * 4 * 3 * x1**2 * x2  # T2
                    + 5 * 3 * 2 * x1 * x2 * x3  # T3
                    + 4 * 2 * 1 * x2 * x3  # T5
                ),
            )
        )
        rcd.append(
            (
                1,
                1,
                (
                    # wrt x2, x2
                    5 * x1 * 2 * x3**2  # T4
                    + 4 * 2 * x3**2  # T6
                    + 3 * 3 * 2 * x2  # T9
                    + 3 * 2 * x3  # T10
                ),
            )
        )
        rcd.append(
            (
                2,
                2,
                (
                    # wrt x3, x3
                    5 * x1 * x2**2 * 2  # T4
                    + 4 * x2**2 * 2  # T6
                    + 4 * 4 * 3 * x3**2  # T7
                ),
            )
        )
        rcd.append(
            (
                1,
                0,
                (
                    # wrt x2, x1
                    5 * 4 * x1**3  # T2
                    + 5 * 3 * x1**2 * x3  # T3
                    + 5 * 2 * x2 * x3**2  # T4
                    + 4 * 2 * x1 * x3  # T5
                    + 3 * x3  # T8
                    + 2  # T11
                ),
            )
        )
        rcd.append(
            (
                2,
                0,
                (
                    # wrt x3, x1
                    5 * 3 * x1**2 * x2  # T3
                    + 5 * x2**2 * 2 * x3  # T4
                    + 4 * 2 * x1 * x2  # T5
                    + 3 * x2  # T8
                ),
            )
        )
        rcd.append(
            (
                2,
                1,
                (
                    # wrt x3, x2
                    5 * x1**3  # T3
                    + 5 * x1 * 2 * x2 * 2 * x3  # T4
                    + 4 * x1**2  # T5
                    + 4 * 2 * x2 * 2 * x3  # T6
                    + 3 * x1  # T8
                    + 3 * 2 * x2  # T10
                    + 2  # T12
                ),
            )
        )

        row = [r for r, _, _ in rcd]
        col = [c for _, c, _ in rcd]
        data = [d for _, _, d in rcd]
        expected_hess = sps.coo_matrix((data, (row, col)), shape=(n_x, n_x))
        expected_hess_array = expected_hess.toarray()
        expected_hess_array = (
            expected_hess_array
            + np.transpose(expected_hess_array)
            - np.diag(np.diagonal(expected_hess_array))
        )
        hess = get_hessian_of_constraint(m.eqn, list(m.x.values()))
        hess_array = hess.toarray()
        np.testing.assert_allclose(expected_hess_array, hess_array, rtol=1e-8)

    def test_unused_variable(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.0)
        m.y = pyo.Var(initialize=1.0)
        m.z = pyo.Var(initialize=1.0)
        m.eqn = pyo.Constraint(expr=m.x**2 + m.y**2 == 1.0)
        variables = [m.x, m.y, m.z]
        expected_hess = np.array([[2, 0, 0], [0, 2, 0], [0, 0, 0]])
        hess = get_hessian_of_constraint(m.eqn, variables).toarray()
        np.testing.assert_allclose(hess, expected_hess, rtol=1e-8)

    def test_explicit_zeros(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.0)
        m.y = pyo.Var(initialize=0.0)
        m.eqn = pyo.Constraint(expr=m.x**2 + m.y**3 == 1.0)
        variables = [m.x, m.y]

        row = np.array([0, 1])
        col = np.array([0, 1])
        data = np.array([2.0, 0.0])
        expected_hess = sps.coo_matrix((data, (row, col)), shape=(2, 2))
        hess = get_hessian_of_constraint(m.eqn, variables)
        np.testing.assert_allclose(hess.row, row, atol=0)
        np.testing.assert_allclose(hess.col, col, atol=0)
        np.testing.assert_allclose(hess.data, data, rtol=1e-8)


@unittest.skipUnless(nx_available, "SCCImplicitFunctionSolver requires networkx")
class TestExternalPyomoModel(unittest.TestCase):
    def test_evaluate_SimpleModel1(self):
        model = SimpleModel1()
        m = model.make_model()
        x_init_list = [[-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]]
        external_model = ExternalPyomoModel(
            [m.x], [m.y], [m.residual_eqn], [m.external_eqn]
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            resid = external_model.evaluate_equality_constraints()
            self.assertAlmostEqual(resid[0], model.evaluate_residual(x[0]), delta=1e-8)

    def test_jacobian_SimpleModel1(self):
        model = SimpleModel1()
        m = model.make_model()
        x_init_list = [[-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]]
        external_model = ExternalPyomoModel(
            [m.x], [m.y], [m.residual_eqn], [m.external_eqn]
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_equality_constraints()
            self.assertAlmostEqual(
                jac.toarray()[0][0], model.evaluate_jacobian(x[0]), delta=1e-8
            )

    def test_hessian_SimpleModel1(self):
        model = SimpleModel1()
        m = model.make_model()
        x_init_list = [[-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]]
        external_model = ExternalPyomoModel(
            [m.x], [m.y], [m.residual_eqn], [m.external_eqn]
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessians_of_residuals()
            self.assertAlmostEqual(
                hess[0][0, 0], model.evaluate_hessian(x[0]), delta=1e-8
            )

    def test_external_hessian_SimpleModel1(self):
        model = SimpleModel1()
        m = model.make_model()
        x_init_list = [[-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]]
        external_model = ExternalPyomoModel(
            [m.x], [m.y], [m.residual_eqn], [m.external_eqn]
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessian_external_variables()
            expected_hess = model.evaluate_external_hessian(x[0])
            self.assertAlmostEqual(hess[0][0, 0], expected_hess, delta=1e-8)

    def test_evaluate_SimpleModel2(self):
        model = SimpleModel2()
        m = model.make_model()
        x_init_list = [[-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]]
        external_model = ExternalPyomoModel(
            [m.x], [m.y], [m.residual_eqn], [m.external_eqn]
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            resid = external_model.evaluate_equality_constraints()
            self.assertAlmostEqual(resid[0], model.evaluate_residual(x[0]), delta=1e-8)

    def test_jacobian_SimpleModel2(self):
        model = SimpleModel2()
        m = model.make_model()
        x_init_list = [[-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]]
        external_model = ExternalPyomoModel(
            [m.x], [m.y], [m.residual_eqn], [m.external_eqn]
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_equality_constraints()
            # evaluate_jacobian_equality_constraints involves an LU
            # factorization and repeated back-solve. SciPy returns a
            # dense matrix from this operation. I am not sure if I should
            # cast it to a sparse matrix. For now it is dense...
            self.assertAlmostEqual(
                jac.toarray()[0][0], model.evaluate_jacobian(x[0]), delta=1e-7
            )

    def test_hessian_SimpleModel2(self):
        model = SimpleModel2()
        m = model.make_model()
        x_init_list = [[-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]]
        external_model = ExternalPyomoModel(
            [m.x], [m.y], [m.residual_eqn], [m.external_eqn]
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessians_of_residuals()
            self.assertAlmostEqual(
                hess[0][0, 0], model.evaluate_hessian(x[0]), delta=1e-7
            )

    def test_external_jacobian_SimpleModel2(self):
        model = SimpleModel2()
        m = model.make_model()
        x_init_list = [[-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]]
        external_model = ExternalPyomoModel(
            [m.x], [m.y], [m.residual_eqn], [m.external_eqn]
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_external_variables()
            expected_jac = model.evaluate_external_jacobian(x[0])
            self.assertAlmostEqual(jac[0, 0], expected_jac, delta=1e-8)

    def test_external_hessian_SimpleModel2(self):
        model = SimpleModel2()
        m = model.make_model()
        x_init_list = [[-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]]
        external_model = ExternalPyomoModel(
            [m.x], [m.y], [m.residual_eqn], [m.external_eqn]
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessian_external_variables()
            expected_hess = model.evaluate_external_hessian(x[0])
            self.assertAlmostEqual(hess[0][0, 0], expected_hess, delta=1e-7)

    def test_external_jacobian_Model2by2(self):
        model = Model2by2()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [0.5, 1.0, 1.5, 2.5, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_external_variables()
            expected_jac = model.evaluate_external_jacobian(x)
            np.testing.assert_allclose(jac, expected_jac, rtol=1e-8)

    def test_external_hessian_Model2by2(self):
        model = Model2by2()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [0.5, 1.0, 1.5, 2.5, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessian_external_variables()
            expected_hess = model.evaluate_external_hessian(x)
            for matrix1, matrix2 in zip(hess, expected_hess):
                np.testing.assert_allclose(matrix1, matrix2, rtol=1e-8)

    def test_external_jacobian_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_external_variables()
            expected_jac = model.evaluate_external_jacobian(x)
            np.testing.assert_allclose(jac, expected_jac, rtol=1e-8)

    def test_external_hessian_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessian_external_variables()
            expected_hess = model.evaluate_external_hessian(x)
            for matrix1, matrix2 in zip(hess, expected_hess):
                np.testing.assert_allclose(matrix1, matrix2, rtol=1e-8)

    def test_evaluate_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            resid = external_model.evaluate_equality_constraints()
            expected_resid = model.evaluate_residual(x)
            np.testing.assert_allclose(resid, expected_resid, rtol=1e-8)

    def test_jacobian_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_equality_constraints()
            expected_jac = model.evaluate_jacobian(x)
            np.testing.assert_allclose(
                jac.toarray(), expected_jac, rtol=1e-8, atol=1e-8
            )

    def test_hessian_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessians_of_residuals()
            expected_hess = model.evaluate_hessian(x)
            np.testing.assert_allclose(hess, expected_hess, rtol=1e-8)

    def test_evaluate_hessian_lagrangian_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x in x_init_list:
            external_model.set_input_values(x)
            external_model.set_equality_constraint_multipliers([1.0, 1.0])
            hess_lag = external_model.evaluate_hessian_equality_constraints()
            hess_lag = hess_lag.toarray()
            expected_hess = model.evaluate_hessian(x)
            expected_hess_lag = np.tril(expected_hess[0] + expected_hess[1])
            np.testing.assert_allclose(hess_lag, expected_hess_lag, rtol=1e-8)


@unittest.skipUnless(nx_available, "SCCImplicitFunctionSolver requires networkx")
class TestUpdatedHessianCalculationMethods(unittest.TestCase):
    """
    These tests exercise the methods for fast Hessian-of-Lagrangian
    computation.
    They use Model2by2 because it has constraints that are nonlinear
    in both x and y.

    """

    def test_external_multipliers_from_residual_multipliers(self):
        model = Model2by2()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [0.5, 1.0, 1.5, 2.5, 4.1]
        lam_init_list = [-2.5, -0.5, 0.0, 1.0, 2.0]
        init_list = list(itertools.product(x0_init_list, x1_init_list, lam_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x0, x1, lam in init_list:
            x = [x0, x1]
            lam = [lam]
            external_model.set_input_values(x)
            lam_g = external_model.calculate_external_constraint_multipliers(lam)
            pred_lam_g = model.calculate_external_multipliers(lam, x)
            np.testing.assert_allclose(lam_g, pred_lam_g, rtol=1e-8)

    def test_full_space_lagrangian_hessians(self):
        model = Model2by2()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [0.5, 1.0, 1.5, 2.5, 4.1]
        lam_init_list = [-2.5, -0.5, 0.0, 1.0, 2.0]
        init_list = list(itertools.product(x0_init_list, x1_init_list, lam_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x0, x1, lam in init_list:
            x = [x0, x1]
            lam = [lam]
            external_model.set_input_values(x)
            # Note that these multiplier calculations are dependent on x,
            # so if we switch their order, we will get "wrong" answers.
            # (This is wrong in the sense that the residual and external
            # multipliers won't necessarily correspond).
            external_model.set_external_constraint_multipliers(lam)
            hlxx, hlxy, hlyy = external_model.get_full_space_lagrangian_hessians()
            pred_hlxx, pred_hlxy, pred_hlyy = (
                model.calculate_full_space_lagrangian_hessians(lam, x)
            )

            # TODO: Is comparing the array representation sufficient here?
            # Should I make sure I get the sparse representation I expect?
            np.testing.assert_allclose(hlxx.toarray(), pred_hlxx, rtol=1e-8)
            np.testing.assert_allclose(hlxy.toarray(), pred_hlxy, rtol=1e-8)
            np.testing.assert_allclose(hlyy.toarray(), pred_hlyy, rtol=1e-8)

    def test_reduced_hessian_lagrangian(self):
        model = Model2by2()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [0.5, 1.0, 1.5, 2.5, 4.1]
        lam_init_list = [-2.5, -0.5, 0.0, 1.0, 2.0]
        init_list = list(itertools.product(x0_init_list, x1_init_list, lam_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x0, x1, lam in init_list:
            x = [x0, x1]
            lam = [lam]
            external_model.set_input_values(x)
            # Same comment as previous test regarding calculation order
            external_model.set_external_constraint_multipliers(lam)
            hlxx, hlxy, hlyy = external_model.get_full_space_lagrangian_hessians()
            hess = external_model.calculate_reduced_hessian_lagrangian(hlxx, hlxy, hlyy)
            pred_hess = model.calculate_reduced_lagrangian_hessian(lam, x)
            # This test asserts that we are doing the block reduction properly.
            np.testing.assert_allclose(np.array(hess), pred_hess, rtol=1e-8)

            from_individual = external_model.evaluate_hessians_of_residuals()
            hl_from_individual = sum(l * h for l, h in zip(lam, from_individual))
            # This test asserts that the block reduction is correct.
            np.testing.assert_allclose(np.array(hess), hl_from_individual, rtol=1e-8)

    def test_evaluate_hessian_equality_constraints(self):
        model = Model2by2()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [0.5, 1.0, 1.5, 2.5, 4.1]
        lam_init_list = [-2.5, -0.5, 0.0, 1.0, 2.0]
        init_list = list(itertools.product(x0_init_list, x1_init_list, lam_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x0, x1, lam in init_list:
            x = [x0, x1]
            lam = [lam]
            external_model.set_input_values(x)
            external_model.set_equality_constraint_multipliers(lam)
            hess = external_model.evaluate_hessian_equality_constraints()
            pred_hess = model.calculate_reduced_lagrangian_hessian(lam, x)
            # This test asserts that we are doing the block reduction properly.
            np.testing.assert_allclose(hess.toarray(), np.tril(pred_hess), rtol=1e-8)

            from_individual = external_model.evaluate_hessians_of_residuals()
            hl_from_individual = sum(l * h for l, h in zip(lam, from_individual))
            # This test asserts that the block reduction is correct.
            np.testing.assert_allclose(
                hess.toarray(), np.tril(hl_from_individual), rtol=1e-8
            )

    def test_evaluate_hessian_equality_constraints_order(self):
        model = Model2by2()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [0.5, 1.0, 1.5, 2.5, 4.1]
        lam_init_list = [-2.5, -0.5, 0.0, 1.0, 2.0]
        init_list = list(itertools.product(x0_init_list, x1_init_list, lam_init_list))
        external_model = ExternalPyomoModel(
            list(m.x.values()),
            list(m.y.values()),
            list(m.residual_eqn.values()),
            list(m.external_eqn.values()),
        )

        for x0, x1, lam in init_list:
            x = [x0, x1]
            lam = [lam]
            external_model.set_equality_constraint_multipliers(lam)
            external_model.set_input_values(x)
            # Using evaluate_hessian_equality_constraints, which calculates
            # external multiplier values, we can calculate the correct Hessian
            # regardless of the order in which primal and dual variables are
            # set.
            hess = external_model.evaluate_hessian_equality_constraints()
            pred_hess = model.calculate_reduced_lagrangian_hessian(lam, x)
            # This test asserts that we are doing the block reduction properly.
            np.testing.assert_allclose(hess.toarray(), np.tril(pred_hess), rtol=1e-8)

            from_individual = external_model.evaluate_hessians_of_residuals()
            hl_from_individual = sum(l * h for l, h in zip(lam, from_individual))
            # This test asserts that the block reduction is correct.
            np.testing.assert_allclose(
                hess.toarray(), np.tril(hl_from_individual), rtol=1e-8
            )


@unittest.skipUnless(nx_available, "SCCImplicitFunctionSolver requires networkx")
class TestScaling(unittest.TestCase):
    def con_3_body(self, x, y, u, v):
        return 1e5 * x**2 + 1e4 * y**2 + 1e1 * u**2 + 1e0 * v**2

    def con_3_rhs(self):
        return 2.0e4

    def con_4_body(self, x, y, u, v):
        return 1e-2 * x + 1e-3 * y + 1e-4 * u + 1e-4 * v

    def con_4_rhs(self):
        return 3.0e-4

    def make_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.0)
        m.y = pyo.Var(initialize=1.0)
        m.u = pyo.Var(initialize=1.0)
        m.v = pyo.Var(initialize=1.0)
        m.con_1 = pyo.Constraint(expr=m.x * m.y == m.u)
        m.con_2 = pyo.Constraint(expr=m.x**2 * m.y**3 == m.v)
        m.con_3 = pyo.Constraint(
            expr=self.con_3_body(m.x, m.y, m.u, m.v) == self.con_3_rhs()
        )
        m.con_4 = pyo.Constraint(
            expr=self.con_4_body(m.x, m.y, m.u, m.v) == self.con_4_rhs()
        )

        epm_model = pyo.ConcreteModel()
        epm_model.x = pyo.Reference(m.x)
        epm_model.y = pyo.Reference(m.y)
        epm_model.u = pyo.Reference(m.u)
        epm_model.v = pyo.Reference(m.v)
        epm_model.epm = ExternalPyomoModel(
            [m.u, m.v], [m.x, m.y], [m.con_3, m.con_4], [m.con_1, m.con_2]
        )
        epm_model.obj = pyo.Objective(expr=m.x**2 + m.y**2 + m.u**2 + m.v**2)
        epm_model.egb = ExternalGreyBoxBlock()
        epm_model.egb.set_external_model(epm_model.epm, inputs=[m.u, m.v])
        return epm_model

    def test_get_set_scaling_factors(self):
        m = self.make_model()
        scaling_factors = [1e-4, 1e4]
        m.epm.set_equality_constraint_scaling_factors(scaling_factors)
        epm_sf = m.epm.get_equality_constraint_scaling_factors()
        np.testing.assert_array_equal(scaling_factors, epm_sf)

    def test_pyomo_nlp(self):
        m = self.make_model()
        scaling_factors = [1e-4, 1e4]
        m.epm.set_equality_constraint_scaling_factors(scaling_factors)
        nlp = PyomoNLPWithGreyBoxBlocks(m)
        nlp_sf = nlp.get_constraints_scaling()
        np.testing.assert_array_equal(scaling_factors, nlp_sf)

    @unittest.skipUnless(cyipopt_available, "cyipopt is not available")
    def test_cyipopt_nlp(self):
        m = self.make_model()
        scaling_factors = [1e-4, 1e4]
        m.epm.set_equality_constraint_scaling_factors(scaling_factors)
        nlp = PyomoNLPWithGreyBoxBlocks(m)
        cyipopt_nlp = CyIpoptNLP(nlp)
        obj_scaling, x_scaling, g_scaling = cyipopt_nlp.scaling_factors()
        np.testing.assert_array_equal(scaling_factors, g_scaling)

    @unittest.skipUnless(cyipopt_available, "cyipopt is not available")
    def test_cyipopt_callback(self):
        # Use a callback to check that the reported infeasibility is
        # due to the scaled equality constraints.
        # Note that the scaled infeasibility is not what we see if we
        # call solve with tee=True, as by default the displayed infeasibility
        # is unscaled. Luckily, we can still access the scaled infeasibility
        # with a callback.
        m = self.make_model()
        scaling_factors = [1e-4, 1e4]
        m.epm.set_equality_constraint_scaling_factors(scaling_factors)
        nlp = PyomoNLPWithGreyBoxBlocks(m)

        def callback(
            local_nlp,
            alg_mod,
            iter_count,
            obj_value,
            inf_pr,
            inf_du,
            mu,
            d_norm,
            regularization_size,
            alpha_du,
            alpha_pr,
            ls_trials,
        ):
            primals = tuple(local_nlp.get_primals())
            # I happen to know the order of the primals here
            u, v, x, y = primals

            # Calculate the scaled residuals I expect
            con_3_resid = scaling_factors[0] * abs(
                self.con_3_body(x, y, u, v) - self.con_3_rhs()
            )
            con_4_resid = scaling_factors[1] * abs(
                self.con_4_body(x, y, u, v) - self.con_4_rhs()
            )
            pred_inf_pr = max(con_3_resid, con_4_resid)

            # Make sure Ipopt is using the scaled constraints internally
            self.assertAlmostEqual(inf_pr, pred_inf_pr)

        cyipopt_nlp = CyIpoptNLP(nlp, intermediate_callback=callback)
        x0 = nlp.get_primals()
        cyipopt = CyIpoptSolver(
            cyipopt_nlp, options={"max_iter": 0, "nlp_scaling_method": "user-scaling"}
        )
        cyipopt.solve(x0=x0)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pickle

import pyomo.common.unittest as unittest

from pyomo.common.dependencies import dill, dill_available as has_dill
from pyomo.core.expr.numvalue import (
    NumericValue,
    is_fixed,
    is_constant,
    is_potentially_variable,
)
from pyomo.kernel import pprint
from pyomo.core.tests.unit.kernel.test_dict_container import (
    _TestActiveDictContainerBase,
)
from pyomo.core.tests.unit.kernel.test_tuple_container import (
    _TestActiveTupleContainerBase,
)
from pyomo.core.tests.unit.kernel.test_list_container import (
    _TestActiveListContainerBase,
)
from pyomo.core.kernel.base import ICategorizedObject
from pyomo.core.kernel.parameter import (
    IParameter,
    parameter,
    functional_value,
    parameter_dict,
    parameter_tuple,
    parameter_list,
)
from pyomo.core.kernel.variable import variable
from pyomo.core.kernel.block import block


class Test_parameter(unittest.TestCase):
    def test_pprint(self):
        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        p = parameter()
        pprint(p)
        b = block()
        b.p = p
        pprint(p)
        pprint(b)
        m = block()
        m.b = b
        pprint(p)
        pprint(b)
        pprint(m)

    def test_ctype(self):
        p = parameter()
        self.assertIs(p.ctype, IParameter)
        self.assertIs(type(p), parameter)
        self.assertIs(type(p)._ctype, IParameter)

    def test_pickle(self):
        p = parameter(value=1.0)
        self.assertEqual(p.value, 1.0)
        self.assertIs(p.parent, None)
        pup = pickle.loads(pickle.dumps(p))
        self.assertEqual(pup.value, 1.0)
        self.assertIs(pup.parent, None)
        b = block()
        b.p = p
        self.assertIs(p.parent, b)
        bup = pickle.loads(pickle.dumps(b))
        pup = bup.p
        self.assertEqual(pup.value, 1.0)
        self.assertIs(pup.parent, bup)

    def test_init(self):
        p = parameter()
        self.assertTrue(p.parent is None)
        self.assertEqual(p.ctype, IParameter)
        self.assertEqual(p.value, None)
        self.assertEqual(p(), None)
        p.value = 1
        self.assertEqual(p.value, 1)
        self.assertEqual(p(), 1)

    def test_type(self):
        p = parameter()
        self.assertTrue(isinstance(p, ICategorizedObject))
        self.assertTrue(isinstance(p, IParameter))
        self.assertTrue(isinstance(p, NumericValue))

    def test_is_constant(self):
        p = parameter()
        self.assertEqual(p.is_constant(), False)
        self.assertEqual(is_constant(p), False)
        p.value = 1.0
        self.assertEqual(p.is_constant(), False)
        self.assertEqual(is_constant(p), False)

    def test_is_fixed(self):
        p = parameter()
        self.assertEqual(p.is_fixed(), True)
        self.assertEqual(is_fixed(p), True)
        p.value = 1.0
        self.assertEqual(p.is_fixed(), True)
        self.assertEqual(is_fixed(p), True)

    def test_potentially_variable(self):
        p = parameter()
        self.assertEqual(p.is_potentially_variable(), False)
        self.assertEqual(is_potentially_variable(p), False)
        p.value = 1.0
        self.assertEqual(p.is_potentially_variable(), False)
        self.assertEqual(is_potentially_variable(p), False)

    def test_polynomial_degree(self):
        p = parameter()
        self.assertEqual(p.polynomial_degree(), 0)
        self.assertEqual((p**2).polynomial_degree(), 0)
        self.assertEqual(p.value, None)
        with self.assertRaises(ValueError):
            (p**2)()
        p.value = 1.0
        self.assertEqual(p.polynomial_degree(), 0)
        self.assertEqual((p**2).polynomial_degree(), 0)
        self.assertEqual(p.value, 1.0)
        self.assertEqual((p**2)(), 1.0)

    def test_is_expression_type(self):
        p = parameter()
        self.assertEqual(p.is_expression_type(), False)

    def test_is_parameter_type(self):
        p = parameter()
        # GH: apparently is_parameter_type has something
        #     to do with mutability...
        self.assertEqual(p.is_parameter_type(), False)


class Test_functional_value(unittest.TestCase):
    def test_pprint(self):
        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        f = functional_value()
        pprint(f)
        b = block()
        b.f = f
        pprint(f)
        pprint(b)
        m = block()
        m.b = b
        pprint(f)
        pprint(b)
        pprint(m)

    def test_ctype(self):
        f = functional_value()
        self.assertIs(f.ctype, IParameter)
        self.assertIs(type(f), functional_value)
        self.assertIs(type(f)._ctype, IParameter)

    def test_pickle(self):
        f = functional_value()
        self.assertIs(f.fn, None)
        self.assertIs(f.parent, None)
        fup = pickle.loads(pickle.dumps(f))
        self.assertIs(fup.fn, None)
        self.assertIs(fup.parent, None)
        b = block()
        b.f = f
        self.assertIs(f.parent, b)
        bup = pickle.loads(pickle.dumps(b))
        fup = bup.f
        self.assertIs(fup.fn, None)
        self.assertIs(fup.parent, bup)

    @unittest.skipIf(not has_dill, "The dill module is not available")
    def test_dill(self):
        p = parameter(1)
        f = functional_value(lambda: p())
        self.assertEqual(f(), 1)
        fup = dill.loads(dill.dumps(f))
        p.value = 2
        self.assertEqual(f(), 2)
        self.assertEqual(fup(), 1)
        b = block()
        b.p = p
        b.f = f
        self.assertEqual(b.f(), 2)
        bup = dill.loads(dill.dumps(b))
        fup = bup.f
        b.p.value = 4
        self.assertEqual(b.f(), 4)
        self.assertEqual(bup.f(), 2)
        bup.p.value = 4
        self.assertEqual(bup.f(), 4)

    def test_call(self):
        f = functional_value()
        self.assertEqual(f(), None)
        self.assertIs(f.fn, None)
        f.fn = lambda: variable(value=1)
        self.assertIsNot(f.fn, None)
        # the function did not return numeric data
        # (always causes an exception)
        with self.assertRaises(TypeError):
            f(exception=False)
        with self.assertRaises(TypeError):
            f(exception=True)
        with self.assertRaises(TypeError):
            f()
        f.fn = lambda: None
        self.assertIsNot(f.fn, None)
        # the function did not return numeric data
        # (always causes an exception)
        with self.assertRaises(TypeError):
            f(exception=False)
        with self.assertRaises(TypeError):
            f(exception=True)
        with self.assertRaises(TypeError):
            f()

        def value_error():
            raise ValueError()

        f.fn = value_error
        self.assertIsNot(f.fn, None)
        self.assertEqual(f(exception=False), None)
        with self.assertRaises(ValueError):
            f(exception=True)
        with self.assertRaises(ValueError):
            f()

    def test_init(self):
        f = functional_value()
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IParameter)
        self.assertEqual(f.fn, None)
        self.assertEqual(f(), None)
        x = [1, 2]
        f.fn = lambda: max(x)
        self.assertEqual(f(), 2)
        x[0] = 3
        self.assertEqual(f(), 3)

    def test_type(self):
        f = functional_value()
        self.assertTrue(isinstance(f, ICategorizedObject))
        self.assertTrue(isinstance(f, IParameter))
        self.assertTrue(isinstance(f, NumericValue))

    def test_is_constant(self):
        f = functional_value()
        self.assertEqual(f.is_constant(), False)
        self.assertEqual(is_constant(f), False)
        f.fn = lambda: 2
        self.assertEqual(f.is_constant(), False)
        self.assertEqual(is_constant(f), False)

    def test_is_fixed(self):
        f = functional_value()
        self.assertEqual(f.is_fixed(), True)
        self.assertEqual(is_fixed(f), True)
        f.fn = lambda: 2
        self.assertEqual(f.is_fixed(), True)
        self.assertEqual(is_fixed(f), True)

    def test_potentially_variable(self):
        f = functional_value()
        self.assertEqual(f.is_potentially_variable(), False)
        self.assertEqual(is_potentially_variable(f), False)
        f.fn = lambda: 2
        self.assertEqual(f.is_potentially_variable(), False)
        self.assertEqual(is_potentially_variable(f), False)

    def test_polynomial_degree(self):
        f = functional_value()
        self.assertEqual(f.polynomial_degree(), 0)
        self.assertEqual((f**2).polynomial_degree(), 0)
        self.assertIs(f.fn, None)
        with self.assertRaises(ValueError):
            (f**2)()
        f.fn = lambda: 2
        self.assertEqual(f.polynomial_degree(), 0)
        self.assertEqual((f**2).polynomial_degree(), 0)
        self.assertEqual(f(), 2)
        self.assertEqual((f**2)(), 4)

    def test_is_expression_type(self):
        f = functional_value()
        self.assertEqual(f.is_expression_type(), False)

    def test_is_parameter_type(self):
        f = functional_value()
        # GH: apparently is_parameter_type has something
        #     to do with mutability...
        self.assertEqual(f.is_parameter_type(), False)


class Test_parameter_dict(_TestActiveDictContainerBase, unittest.TestCase):
    _container_type = parameter_dict
    _ctype_factory = lambda self: parameter()


class Test_parameter_tuple(_TestActiveTupleContainerBase, unittest.TestCase):
    _container_type = parameter_tuple
    _ctype_factory = lambda self: parameter()


class Test_parameter_list(_TestActiveListContainerBase, unittest.TestCase):
    _container_type = parameter_list
    _ctype_factory = lambda self: parameter()


if __name__ == "__main__":
    unittest.main()
